"""
Licensed under a 3-clause BSD style license - see LICENSE.rst

This class supports communication with DS9 through the XPA


Some code in this class was adapted from pysao, which can be found
at https://github.com/leejjoon/pysao. Specifically this package used the
existing Cython implementation to the XPA  and extended the calls to the other
available XPA executables so that more functionality is added. The API
information is available here:

http://hea-www.harvard.edu/RD/xpa/client.html#xpaopen

Using Cython will allow for broader development of the code and produce faster
runtimes for large datasets with repeated calls to the display manager.

XPA is licensed under MIT, help can be found here:
http://hea-www.cfa.harvard.edu/saord/xpa/help.html

The current XPA can be downloaded from here:
http://hea-www.harvard.edu/saord/xpa/

"""

import os
import shutil
import sys
import numpy as np
from subprocess import Popen
import time
import warnings
import logging
from tempfile import mkdtemp
import matplotlib.image as mpimage
import matplotlib.pyplot as plt
from matplotlib import get_backend
import atexit
from subprocess import call

# The XPA class controls interaction with DS9
from .xpa_wrap import XPA
from .imexamxpa import XpaException

from . import util
from astropy.io import fits


class UnsupportedDatatypeException(Exception):
    pass


class UnsupportedImageShapeException(Exception):
    pass

__all__ = ['ds9']


class ds9:
    """Control all interactions between the user and the DS9 window.

    The ds9() contructor takes a ds9 target as its main argument.
    If none is given, then a new window and process will be started.

    DS9's xpa access points are documented in the reference manual,
    but the can also be returned to the user with a call to xpaset.

    http://hea-www.harvard.edu/saord/ds9/ref/xpa.html


    Parameters
    ----------
    target: string, optional
         the ds9 target name or id. If None or empty string,
         a new ds9 instance is created.

    path : string, optional
        path of the ds9. Used only if a new ds9 is requested.

    wait_time : float, optional
        waiting time before error is raised

    quit_ds9_on_del : boolean, optional
        If True, try to quit ds9 when this instance is deleted.


    Attributes
    ----------

    wait_time: float
         The waiting time before error is raised

    path: string
        The path to the DS9 executable

    _xpa_name: string
        The value in XPA_METHOD, the name of the DS9 window

    _quit_ds9_on_del: boolean
         Determine whether to quit ds9 when object goes out of scope.

    _ds9_unix_name: string
        The full path filename to the unix socket, only if unix sockets are
        being used with local

    _need_to_purge: boolean
        whenever there are unix socket directories which need to be purged when
        the object goes out of scope

    _tmpd_name: string
        The full path name to the unix socket file on the local system

    _filename: string
        The name of the image that's currently loaded into DS9

    _ext: int
        Extension of the current image that is loaded. If one extension of an
        MEF file is loaded this will be 1 regardless of the extension that was
        specified (because DS9 and the XPA now see it as a single image and
        header)

    _extname: string
        If available, the EXTNAME of the MEF extension that is loaded, taken
        from the current data header

    _extver: int
        If available, the EXTVER of the MEF extension that is loaded, taken
        from the current data header

    _ds9_process: pointer
        Points to the ds9 process id on the system, returned by Popen, whenever
        this module starts DS9

    _mef_file: boolean
        The file is a multi-extension fits file

    _iscube: bookean
        The file is a multiextension fits file, and one of the extensions
        contains at least 1 additional extension (3D or more)

    _numaxis: int
        number of image planes, this is NAXIS

    _naxis: tuple
        specific image plane displayed, defaulted to 1 image plane, most
        relevant to cube fits files
    """

    # _ImgCode : copied from fits, used for displaying arrays straight to DS9
    _ImgCode = {'float32': -32,
                'float64': -64,
                'float16': -16,
                'int16': 16,
                'int32': 32,
                'int64': 64,
                'uint8': 8,
                'uint16': 16}

    _tmp_dir = ""
    _process_list = list()

    def __init__(self, target='', path='', wait_time=5,
                 quit_ds9_on_del=True):
        """starter.

        Notes
        -----
        I think this is a quirk in the XPA communication. The xpans, and XPA
        prefer to have all connections be of the same type. DS9 defaults to
        creating an INET connection. In some cases, if no IP address can be
        found for the computer, the startup can hang. In these cases, a local
        connection is preferred, which uses a unix filename for the socket.

        The problem arises that if the user already has DS9 windows running,
        that were started by default, the nameserver is only listening for the
        default socket type (inet) and not local. There are also cases where
        the machine running this code does not have xpa installed, so there is
        no xpans (nameserver) to run and keep track of the open connections.
        In that case, the user needs to provide this init with the name of the
        socket in their window (in XPA_METHOD) in order to create the
        connection.

        """
        # determine whether to quit ds9 also when object deleted.
        self._quit_ds9_on_del = quit_ds9_on_del
        self.wait_time = wait_time
        self._need_to_purge = False
        self._tmpd_name = None

        # dictionary where each key is a frame number, and the values are a
        # dictionary of details about the image loaded in that frame
        self._viewer = dict()
        self._current_frame = None
        self._current_slice = None

        # default starting socket type to inet
        # users can change to local using environment variable
        self._xpa_method = "inet"
        self._xpa_name = ""

        # only used for DS9 windows started from this module
        self._ds9_process = None
        self._ds9_path = None

        # see if the package logger was already started
        self.log = logging.getLogger(__name__)

        # An existing ds9 was suggested. Confirm existence
        # and attach.
        if target:
            # check to see if the target exists
            active = util.list_active_ds9(False)
            if target in list(active):
                self._xpa_name = target
            else:
                # see if they used a title and not the xpa_method
                # the title is the first item in the tuple for each key
                for window in active.values():
                    if target in window:
                        self._xpa_name = target
            if not self._xpa_name:
                print(f"\nDS9 target: {target} doesn't exist.\n{list(active)}")
                raise ValueError(
                    "Please choose an existing target."
                )

        else:
            if not path:
                self._ds9_path = util.find_path('ds9')
                if not self._ds9_path:
                    raise OSError("Could not find ds9 executable on your path or alias")

            else:
                self._ds9_path = path

            # Check to see if the user has chosen a preference first
            if 'XPA_METHOD' in os.environ:
                self._xpa_method = os.environ['XPA_METHOD'].lower()

            if 'inet' in self._xpa_method:
                self._xpa_name = self.run_inet_ds9()
                # xpa_name is the title of the window, the xpa can be
                # referenced with either the socket address or name
                # of the window

            elif 'local' in self._xpa_method:
                self._xpa_name, self._ds9_unix_name = self._run_unixonly_ds9()

            # tells xpa where to find sockets
            os.environ['XPA_METHOD'] = self._xpa_method

        # xpa_name sets the template for the get and set commands
        self.xpa = XPA(self._xpa_name)
        if 'local' in self._xpa_method:
            self._set_frameinfo()  # initial load
        self._define_cmaps()  # dictionary of color maps

    def __str__(self):
        pass

    def __del__(self):
        """Nicely exit the DS9 process."""
        if self._quit_ds9_on_del:
            if 'local' in self._xpa_method:
                self._purge_local()
            else:
                self._stop_process()

    def _set_frameinfo(self):
        """Set the name and extension for the data displayed in current frame.

        Notes
        -----
        The absolute path reference is stored to make XPA happy in all cases,
        wherever the user started the DS9 process.

        The only consistant way to return which cube and slice that is
        displayed is with the call to "file" which has the full plane=x:y
        information, but only when looking at something other than the first
        extension for each plane. In this case, you have to look at the header
        information to see it's a cube image, and assume the first image plane
        is displayed.

        If you load a single extension from an MEF into DS9, XPA references
        the extension as 1 afterwards for access points you need to look in
        the header of the displayed image to find out what actual extension
        is loaded

        This also also gathers needed image header information
        It's possible the user has an array in the frame, for which there is no
        header or filename information

        """
        # check the current frame, if none exists, then don't continue
        frame = self.frame()
        if frame:

            if frame not in self._viewer:
                self._viewer[frame] = dict()

            try:
                user_array = self._viewer[frame]['user_array']
            except KeyError:
                user_array = None

            extver = None  # extension number
            extname = None  # name of extension
            filename = None  # filename of image
            numaxis = 2  # number of image planes, this is NAXIS
            # tuple of each image plane, defaulted to 1 image plane
            naxis = (0)
            # data has more than 2 dimensions and loads in cube/slice frame
            iscube = False
            mef_file = False  # used to check misleading headers in fits files

            load_header = False  # not used for in memory arrays
            self._current_frame = frame

            # see if any file is currently loaded into ds9,
            # xpa returns '\n' for nothing loaded
            # get the current frame
            try:
                filename_string = self.get('file').strip()
                if len(filename_string) > 1 and '\n' not in filename_string:
                    filename = str(filename_string.strip().split('[')[0])
                    self._viewer[frame]['filename'] = os.path.abspath(filename)
                    load_header = True
                else:
                    filename_string = ""
            except XpaException:
                filename_string = ""

            try:
                if "plane" in filename_string:
                    iscube = True
                    if ":" in filename_string:
                        naxis = filename_string.strip().split(']')[
                            1].split("=")[1].split(":")
                    else:
                        naxis = filename_string.strip().split(']')[
                            1].split('=')[1].split()
                    if len(naxis) == 1:
                        naxis.append("0")
                    naxis.reverse()  # for astropy.fits row-major ordering
                    naxis = map(int, naxis)
                    naxis = [axis - 1 if axis > 0 else 0 for axis in naxis]
                    naxis = tuple(naxis)
            except ValueError:
                raise ValueError("Problem parsing filename")

            if load_header:
                # set the extension from the header information returned from
                # DS9 this is the best way to get the information if the user
                # changes the loaded file using the gui
                header_cards = fits.Header.fromstring(
                    self.get_header(),
                    sep='\n')
                mef_file, nextend, first_image = util.check_valid(filename)
                if mef_file:
                    try:
                        extver = int(header_cards['EXTVER'])
                    except KeyError:
                        # fits doesn't require extver if there is only 1
                        # extension
                        extver = first_image

                    try:
                        extname = str(header_cards['EXTNAME'])
                    except KeyError:
                        extname = None

                try:
                    numaxis = int(header_cards['NAXIS'])
                except KeyError:
                    raise KeyError("Problem getting NAXIS from header")

                if not iscube:
                    if numaxis > 2:
                        iscube = True
                        naxis = list()
                        # assume the first axis in each extension is displayed
                        for axis in range(numaxis, 2, -1):
                            naxis.append(0)
                        naxis = tuple(naxis)

            # update the viewer dictionary, if the user changes what's
            # displayed in a frame this should update correctly
            # this dictionary will be referenced in the other parts of
            # the code. This enables tracking user arrays through
            # frame changes

            self._viewer[frame] = {'filename': filename,
                                   'extver': extver,
                                   'extname': extname,
                                   'naxis': naxis,
                                   'numaxis': numaxis,
                                   'iscube': iscube,
                                   'user_array': user_array,
                                   'mef': mef_file}

        else:
            warnings.warn("No frame loaded in viewer")

    def valid_data_in_viewer(self):
        """return bool if valid file or array is loaded into the viewer."""
        frame = self.frame()

        if frame:
            self._set_frameinfo()
            if self._viewer[frame]['filename']:
                return True
            else:
                try:
                    if self._viewer[frame]['user_array'].any():
                        return True
                except AttributeError as ValueError:
                    print("error in array")
        return False

    def get_filename(self):
        """return the filename currently on display.

        This function will check if there is already a filename saved. It's
        possible that the user can connect to a ds9 window with no file loaded
        and then ask for the data file name after loading one through the ds9
        menu options. This will poll the private filename and then try and set
        one if it's empty.
        """
        # see if the user has loaded a file by hand or changed frames in the
        # gui
        frame = self.frame()
        self._set_frameinfo()
        return self._viewer[frame]['filename']

    def get_frame_info(self):
        """return more explicit information about the data displayed."""
        self._set_frameinfo()
        return self._viewer[self.frame()]

    def get_viewer_info(self):
        """Return a dictionary of information.

        The dictionary contains information about all frames which are
        loaded with data

        Notes
        -----
        Consider adding a loop to verify that all the frames still exist
        and the user has not deleted any through the gui.
        """
        self._set_frameinfo()
        return self._viewer

    @classmethod
    def _purge_tmp_dirs(cls):
        """Delete temporary directories made for the unix socket.

        When used with ipython (pylab mode), it seems that the objects
        are not properly deleted, i.e., temporary directories are not
        deleted. This is a work around for that.
        """

        if cls._tmp_dir:
            shutil.rmtree(cls._tmp_dir)

    @classmethod
    def _stop_running_process(cls):
        """stop self generated DS9 windows when user quits python window."""
        while cls._process_list:
            process = cls._process_list.pop()
            if process.poll() is None:
                process.terminate()

    def _stop_process(self):
        """stop the ds9 window process nicely.

        but only if this package started it
        """
        try:
            if self._ds9_process:
                # none means not yet terminated
                if self._ds9_process.poll() is None:
                    self.set("exit")
                    if self._ds9_process in self._process_list:
                        self._process_list.remove(self._ds9_process)

        except XpaException as e:
            print(f"XPA Exception: {repr(e)}")

    def _purge_local(self):
        """remove temporary directories from the unix socket."""
        if not self._need_to_purge:
            return

        if not self._quit_ds9_on_del:
            warnings.warn(
                "You need to manually delete tmp. dir ({0:s})".format(
                    self._tmpd_name))
            return

        self._stop_process()
        # add a wait for the process to terminate before trying to delete the
        # tree
        time.sleep(0.5)

        try:
            shutil.rmtree(self._tmpd_name)
        except OSError:
            warnings.warn(
                "Warning : couldn't delete the temporary \
                directory ({0:s})".format(self._tmpd_name,))

        self._need_to_purge = False

    def close(self):
        """close the window and end connection."""
        # make sure we clean up the object and quit_ds9 local files
        if 'local' in self._xpa_method or 'tmp' in self._xpa_method:
            self._purge_local()
        else:
            self._stop_process()

    def run_inet_ds9(self):
        """start a new ds9 window using an inet socket connection.

        Notes
        -----
        It is given a unique title so it can be identified later.
        """
        env = os.environ

        # this is the title of the window, without a nameserver connection
        # is there a way to get the inet x:x address?
        # that should be unique enough, something better?
        xpaname = "imexam" + str(time.time())
        try:
            p = Popen([self._ds9_path,
                       "-xpa", "inet",
                       "-title", xpaname],
                      shell=False, env=env)
            self._ds9_process = p
            self._process_list.append(p)
            self._need_to_purge = False
            return xpaname

        except Exception as e:  # refine error class
            warnings.warn("Opening ds9 failed")
            print(f"Exception: {repr(e)}")
            from signal import SIGTERM
            try:
                pidtokill = p.pid
            except NameError:
                # in case p failed at the initialization level
                pidtokill = None
            if pidtokill is not None:
                os.kill(pidtokill, SIGTERM)
            raise e

    def _run_unixonly_ds9(self):
        """start new ds9 window and connect to object using a unix socket.

        Notes
        -----
        When the xpa method in libxpa parses a given template as a unix
        socket, it checks if the template string starts with tmpdir
        (from env["XPA_TMPDIR"] or default to /tmp/.xpa). This can make
        having multiple instances of ds9 a bit difficult, but if you give it
        unique names or use the inet address you should be fine

        For unix only, we run ds9 with XPA_TMPDIR set to temporary directory
        whose prefix start with /tmp/xpa (eg, /tmp/xpa_sf23f), them set
        os.environ["XPA_TMPDIR"] (which affects xpa set and/or get command
        from python) to /tmp/xpa.
        """
        env = os.environ
        wait_time = self.wait_time

        self._tmpd_name = mkdtemp(
            prefix="xpa_" +
            env.get(
                "USER",
                ""),
            dir="/tmp")

        # this is the first directory the servers looks for on the path
        env["XPA_TMPDIR"] = self._tmpd_name

        unix_name = "{0:s}/.IMT".format(self._tmpd_name)

        # that should be unique enough, something better?
        title = str(time.time())
        try:
            # unix only flag disables the fifos and inet connections
            p = Popen([self._ds9_path,
                       "-xpa", "local",
                       "-unix_only", "-title", title,
                       "-unix", "{0:s}".format(unix_name)],
                      shell=False, env=env)

            # wait until ds9 starts and the .IMT socket exists
            while wait_time > 0:
                file_list = os.listdir(self._tmpd_name)
                if ".IMT" in file_list:
                    break
                time.sleep(0.5)
                wait_time -= 0.5

            if wait_time == 0:
                from signal import SIGTERM
                os.kill(p.pid, SIGTERM)
                print(f"Connection timeout with the ds9. Try to increase the \
                    *wait_time* parameter (current value \
                    is  {self.wait_time} s)")

        except (OSError, ValueError, AttributeError) as e:
            warnings.warn("Starting ds9 failed")
            shutil.rmtree(self._tmpd_name)

        else:
            self._tmp_dir = self._tmpd_name
            self._ds9_process = p
            self._process_list.append(p)

        # this might be sketchy
        try:
            file_list.remove(".IMT")  # should be in the directory, if not
        except (ValueError, IOError):
            warnings.warn("IMT not found in tmp, using first thing in list")
        if len(file_list) > 0:
            xpaname = os.path.join(self._tmpd_name, file_list[0])
        else:
            shutil.rmtree(self._tmpd_name)
            raise ValueError("Problem starting ds9 local socket connection")

        env["XPA_TMPDIR"] = "/tmp/xpa"  # for all local connections
        self._need_to_purge = True
        self._xpa_method = 'local'
        return xpaname, unix_name

    def _check_ds9_process(self):
        """Check to see if the ds9 process is still running.

        Notes
        -----
        If you start a ds9 window from the shell and then connect
        to imexam, imexam will not have a reference for the process,
        so this method ignores that state.
        """
        if self._ds9_process:
            ret = self._ds9_process.poll()
            if ret:
                raise RuntimeError("The ds9 process is externally killed.")
                self._purge_local()

    def set(self, param, buf=None):
        """XPA set method to ds9 instance.

        Notes
        -----
        This function is linked with the Cython implementation

        set(param, buf=None)
        param : parameter string (eg. "fits" "regions")
        buf : aux data string (sometime string needed to be ended with CR)
        """
        self._check_ds9_process()
        self.xpa.set(param, buf)

    def get(self, param):
        """XPA get method to ds9 instance which returns received string.

        Parameters
        ----------
        param : parameter string (eg. "fits" "regions")


        Notes
        -----
        This function is linked with the Cython implementation
        get(param)

        """
        self._check_ds9_process()
        return self.xpa.get(param)

    def readcursor(self):
        """Returns the image coordinate postion and key pressed.

        Notes
        -----
        XPA returns strings of the form:  u a  257.5 239
        """
        try:
            xpa_string = self.get("imexam any coordinate image")
        except XpaException as e:
            print(f"Xpa problem reading cursor: {repr(e)}")
            raise KeyError
        except ValueError:
            raise ValueError("Outside of data range")

        k, x, y = xpa_string.split()

        # ds9 is returning 1 based array, set to 0-based since
        # imexamine uses numpy for array crunching
        return float(x)-1, float(y)-1, str(k)

    def alignwcs(self, on=True):
        """align wcs.

        Parameters
        ----------
        on: bool
            Align the images using the WCS in their headers
        """
        self.set("align {0:s}".format(str(on)))

    def blink(self, blink=True, interval=None):
        """Blink frames.

        Parameters
        ----------
        blink: bool, optional
            Set to True to start blinking the frames which are open

        interval: int
            Set interval equal to the length of pause for blinking


        Notes
        -----
        blink_syntax=
            Syntax:
            blink
            [true|false]
            [interval <value>]

        """
        cstring = "blink "
        if blink:
            cstring += "yes "
        else:
            cstring += "no "
            if interval:
                cstring += " {0:d}".format(interval)

        self.set(cstring)

    def clear_contour(self):
        """clear contours from the screen."""
        self.set("contour clear")

    def _define_cmaps(self):
        """setup the default color maps which are available."""
        self._cmap_syntax = """
        Syntax:
        cmap [<colormap>]
            [file]
            [load <filename>]
            [save <filename>]
            [invert true|false]
            [value <constrast> <bias>]
            [tag [load|save] <filename>]
            [tag delete]
            [match]
            [lock [true|false]]
            [open|close]

         Examples
         --------
            >obj.cmap(map="gist_heat")
            >obj.cmap(invert=True)
         """

        # is there a list I can pull automatically from ds9?
        self._cmap_colors = ["grey", "red", "green", "red", "blue", "a", "b",
                             "bb", "he", "i8", "aips0", "sls", "hsv", "heat",
                             "cool", "rainbow", "standard", "staircase",
                             "color"]

    def cmap(self, color=None, load=None, invert=False, save=False,
             filename='colormap.ds9'):
        """Set the color map table, using a defined list of options.

        Parameters
        ----------
        color: string
            color must be set to one of the available DS9 color map names

        load: string, optional
            set to the filename which is a valid colormap lookup table
            valid contrast values are from 0 to 10, and valid bias
            values are from 0 to 1

        invert: bool, optional
            invert the colormap

        save: bool, optional
            save the current colormap as a file

        filename: string, optional
            the name of the file to save the colormap to
        """
        if color:
            color = color.lower()
            if color.lower() in self._cmap_colors:
                cstring = "cmap {0:s}".format(color)
                self.set(cstring)
            else:
                print("Unrecognized color map, choose one of these:")
                print(self._cmap_colors)

        if invert:
            invert = 'yes'
            cstring = 'cmap invert {0:s}'.format(invert)
            self.set(cstring)

        if load:
            cstring = 'cmap load {0:s}'.format(load)
            self.set(cstring)

        if save:
            cstring = 'cmap save {0:s}'.format(filename)
            self.set(cstring)

        if not color and not load:
            print(self._cmap_colors)

    def colorbar(self, on=True):
        """turn the colorbar on the bottom of the window on and off.

        Parameters
        ----------
        on: bool
            Set to True to turn on the colorbar, False to turn it off

        """
        self.set("colorbar {0:s}".format(str(on)))

    def contour(self, on=True, construct=True):
        """show contours on the window.

        Parameters
        ----------
        on: bool
            Set to true to turn on contours

        construct: bool, optional
            Will open the contour dialog box which has more options

        """
        self.set("contour {0:s}".format(str(on)))
        if construct:
            self.set("contour levels")

    def contour_load(self, filename):
        """load a contour file into the window.

        Parameters
        ----------
        filename: string
            The name of the file with the contour level defined

        """
        if filename:
            self.set("contour loadlevels {0:s}".format(str(filename)))
        else:
            warnings.warn("No filename provided for contours")

    def crosshair(self, x=None, y=None, coordsys="physical",
                  skyframe="wcs", skyformat="fk5", match=False, lock=False):
        """Control the position of the crosshair in the current frame.

        crosshair mode is turned on automatically

        Parameters
        ----------

        x: string or int
            The value of x is converted to a string for the call to XPA,
            use a value here appropriate for the skyformat you choose

        y: string or int
            The value of y is converted to a string for the call to XPA,
            use a value here appropriate for the skyformat you choose

        coordsys: string, optional
            The coordinate system your x and y are defined in

        skyframe: string, optional
            If skyframe has "wcs" in it then skyformat is also sent to the XPA

        skyformat: string, optional
            Used with skyframe, specifies the format of the coordinate
            which were given in x and y

        match: bool, optional
            If set to True, then the wcs is matched for the frames

        lock: bool, optional
            If set to True, then the frame is locked in wcs
        """
        if x and y:
            if "wcs" in skyframe:
                format = skyformat
            else:
                format = ""
            self.window.set("crosshair {0:s} {1:s} {2:s} {3:s}".format(
                str(x), str(y), str(coordsys), format))
        if match:
            self.window.set("crosshair match wcs")
        if lock:
            self.window.set("crosshair lock wcs")

    def cursor(self, x=None, y=None):
        """move the cursor in the current frame to the specified image pixel.

         selected regions will also be moved

        Parameters
        ----------
        x: int
            pixel location  x coordinate to move to

        y: int
            pixel location  y coordinate to move to

        x and y are converted to strings for the call

        """
        if x and y:
            self.set("cursor {0:s} {1:s}".format(str(x), str(y)))
        else:
            warnings.warn(
                "You need to supply both an x and y location for the cursor")

    def disp_header(self):
        """Display the header of the current image to a DS9 window."""
        cstring = "header "
        try:
            self.set(cstring)  # display the header of the current frame
        except XpaException as e:
            raise XpaException("XPA Exception getting header: {0}".format(e))

    def frame(self, n=None):
        """convenience function to change or report frames.

        Parameters
        ----------
        n: int, string, optional
            The frame number to open or change to. If the number specified
            doesn't exist, a new frame will be opened
            If nothing is specified, then the current frame number will be
            returned. The value of n is converted to a string before passing
            to the XPA

        Examples
        --------
        frame(1)  sets the current frame to 1
        frame("last") set the current frame to the last frame
        frame() returns the number of the current frame
        frame("new") opens a new frame
        frame(3)  opens frame 3 if it doesn't exist already, otherwise
        goes to frame 3

        """
        frame = self.get("frame").strip()  # xpa returns '\n' for no frame

        if frame:
            if n:
                if "delete" in str(n):
                    if frame in self._viewer:
                        del self._viewer[frame]
                self.set("frame {0:s}".format(str(n)))
            else:
                try:
                    if len(frame) < 1:
                        self.set("frame 1")
                        # the user can delete frame 1, but ds9 defaults to 1,
                        # so set it
                        return '1'
                    else:
                        return str(frame)
                except XpaException as e:
                    raise XpaException(
                        "XPA Exception getting frame: {0}".format(e))
        else:
            if len(frame) < 1:
                self.set("frame 1")
                # the user can delete frame 1, but ds9 defaults to 1, so set it
                return '1'

    def iscube(self):
        """return whether a cube image is displayed in the current frame."""
        frame = self.frame()
        if frame != self._current_frame:
            self._set_frameinfo()
        else:
            return self._viewer[frame]['iscube']

    def get_slice_info(self):
        """return the slice tuple that is currently displayed."""
        self._set_frameinfo()
        frame = self.frame()

        if self._viewer[frame]['iscube']:
            image_slice = self._viewer[frame]['naxis']
        else:
            image_slice = None
        return image_slice

    def get_data(self):
        """return a numpy array of the data displayed in the current frame.

        Notes
        -----
        This is the data array that the imexam() function from connect()
        uses for analysis

        astropy.io.fits stores data in row-major format. So a 4d image would be
        [NAXIS4, NAXIS3, NAXIS2, NAXIS1]
        just the one image is retured in the case of multidimensional data, not
        the cube

        """
        # make sure the filename and extension info are correct for the current
        # frame in DS9, users can change this in the gui
        # users can change frame and slice before calling this, best to check
        self._set_frameinfo()
        frame = self.frame()
        if frame:

            if isinstance(self._viewer[frame]['user_array'], np.ndarray):
                return self._viewer[frame]['user_array']
            else:
                filename = self._viewer[frame]['filename']
                extver = self._viewer[frame]['extver']
                extname = self._viewer[frame]['extname']
                naxis = self._viewer[frame]['naxis']

                if self._viewer[frame]['mef']:
                    with fits.open(filename) as filedata:
                        if self._viewer[frame]['iscube']:
                            data = filedata[extname, extver].section[naxis]
                        else:
                            data = filedata[extver].data
                    return data
                else:
                    with fits.open(filename) as filedata:
                        data = filedata[0].data
                    return data

    def get_image(self):
        """return the full image object instead of just the data array."""
        print("Returning just the data array, open the image file \
               for the full object")
        self.get_data()

    def get_header(self, fitsobj=False):
        """Return the current fits header.

        The return value is the string or None if there's a problem
        If fits is True  then a fits header object is returned instead
        """
        # TODO return the simple header for arrays which are loaded

        frame = self.frame()
        if frame != self._current_frame:
            self._set_frameinfo()

        if self._viewer[frame]['filename']:
            if fitsobj:
                header = fits.getheader(self._viewer[frame]['filename'])
            else:
                try:
                    header = self.get("fits header")
                except XpaException as e:
                    print(f"XPA Exception getting header: {repr(e)}")
                    return None

            return header
        else:
            warnings.warn("No file with header loaded into ds9")
            return None

    def grid(self, on=True, param=False):
        """convenience to turn the grid on and off.

        grid can be flushed with many more options

        Parameters
        ----------

        on: bool, optional
            Will turn the grid overlay on in the current frame

        param: bool, optional
            Will open the parameter dialog in DS9

        """
        self.set("grid {0:s}".format(str(on)))
        if param:
            self.set("grid open")

    def hideme(self):
        """lower the ds9 window."""
        self.set("lower")

    def embed(self):
        """Embed the viewer in a notebook."""
        print("Not Implemented for DS9")

    def load_fits(self, fname, extver=None, mecube=False):
        """convenience function to load fits image to current frame.

        Parameters
        ----------
        fname: string, FITS object
            The name of the file to be loaded. You can specify the full
            extension in the name, such as
            filename_flt.fits or filename_flt.fits[1]
            You can also pass it an in-memory FITS object

        extver: int, optional
            The extension to load (EXTVER in the header)

        mecube: bool, optional
            If mecube is True, load the fits file as a cube into ds9

        Notes
        -----
        To tell ds9 to open a file whose name or path includes spaces,
        surround the path with "{...}", e.g.

        % xpaset -p ds9 file "{foo bar/my image.fits}"

        This is assuming that the image loads into the current or next
        new frame, watch the internal file and ext values because the user
        can switch frames through DS9 app itself

        XPA needs to have the absolute path to the filename so that if the
        DS9 window was started in another directory it can still find the
        file to load. The pathname also needs to be stripped of spaces.
        """
        # for the viewer reference
        frame = self.frame()
        if frame is None:
            frame = 1  # load into first frame

        if isinstance(fname, fits.hdu.image.PrimaryHDU):
            shortname = fname
            extn = None
            extv = 0
            if extver is None:
                extver = extv
        if isinstance(fname, fits.hdu.hdulist.HDUList):
            shortname = fname
            extn = None
            extv = extver
        elif isinstance(fname, str):
            shortname, extn, extv = util.verify_filename(fname)
            if extn is not None:
                raise ValueError("Extension name  given, must "
                                 "specify the absolute extension you want")
            # prefer the keyword value over the extension in the name
            if extver is None:
                extver = extv
        else:
            raise TypeError("Expected FITS data as input")

        # safety for a valid imexam file
        if ((extv is None) and (extver is None)):
            mef_file, nextend, first_image = util.check_valid(shortname)
            extver = first_image  # the extension of the first IMAGE

        if isinstance(fname, str):
            if mecube:
                cstring = "mecube {0:s}".format(shortname)
            else:
                cstring = ('fits {0:s}[{1:d}]'.format(shortname, extver))
            self.set(cstring)
            # make sure any previous reference is reset
            self._set_frameinfo()
            self._viewer[frame]['user_array'] = None
        else:
            self.view(fname[extver].data)

    def load_region(self, filename):
        """Load regions from a file which uses ds9 standard formatting.

        Parameters
        ----------

        filename: string
            The file containing the DS9 style region description

        """
        if os.access(filename, os.F_OK):
            self.set("regions load {0:s}".format(filename))
        else:
            warnings.warn("No such file:{0:s}".format(filename))

    def load_rgb(self, red, green, blue, scale=False, lockwcs=False):
        """load 3 images into an RGBimage frame.

        Parameters
        ----------
        red: string
            The name of the fits file loaded into the red channel

        green: string
            The name of the fits file loaded into the green channel

        blue: string
            The name of the fits file loaded into the blue channel

        scale: bool
            If True, then each image will be scaled with zscale() after loading

        lockwcs: bool
            If True, then the image positions will be locked to each other
            using the WCS information in their headers

        """
        self.set("rgb new")
        self.set("rgb channel red")
        self.load_fits(red)
        if scale:
            self.scale()
        self.set("rgb channel green")
        self.load_fits(green)
        if scale:
            self.scale()
        self.set("rgb channel blue")
        self.load_fits(blue)
        if scale:
            self.scale()
        if lockwcs:
            self.set("rgb wcs yes")

    def load_mef_as_cube(self, filename=None):
        """Load a Mult-Extension-Fits image into one frame as an image cube."""
        if not filename:
            print("No filename specified")
        else:
            self.set(f"file mecube new {filename}")

    def load_mef_as_multi(self, filename=None):
        """Load a Mult-Extension-Fits image into multiple frames."""
        if not filename:
            print("No filename specified")
        else:
            self.set(f"file multiframe {filename}")

    def mark_region_from_array(self, input_points,
                               ptype="image", textoff=10, size=4):
        """mark ds9 regions regions  given an input list of tuples.

         a convienence function, you can also use set_region

        Parameters
        ----------
        input_points: iterator, a tuple, or list of tuples
                      or a string: (x,y,comment),

        ptype: string
            the reference system for the point locations, image|physical|fk5

        size: int
            the size of the region marker

        textoff: string
            the offset for the comment text, if comment is empty
            it will not show

        Notes
        -----
        only circular regions are supported currently

        """
        if isinstance(input_points, tuple):
            input_points = [input_points]
        elif isinstance(input_points, str):
            input_points = [tuple(input_points.split())]

        x = 0
        y = 1
        comment = 2
        rtype = "circle"  # only one supported right now

        for location in input_points:
            if rtype == "circle":
                pline = rtype + " " + \
                    str(location[x]) + " " + str(location[y]) + " " + str(size)
                self.log.info(pline)
                self.set_region(pline)

            try:
                if(len(str(location[comment])) > 0):
                    pline = "text " + str(float(location[x]) + textoff) +\
                            " " + str(float(location[y]) + textoff) + " '" +\
                            str(location[comment]) + "' #font=times"
                    self.log.info(pline)
                    self.set_region(pline)
            except IndexError:
                pass

    def make_region(self, infile, labels=False, header=0, textoff=10, size=5):
        """make an input reg file with  [x,y,comment] to a DS9 reg file.

        the input file should contain lines specifying x,y,comment

        Parameters
        ----------

        infile: str
            input filename

        labels: bool
            add labels to the regions

        header: int
            number of header lines in text file to skip

        textoff: int
            offset in pixels for labels

        size: int
            size of the region type


        Notes
        -----
        only circular regions are supported currently
        """
        try:
            f = open(infile, 'r')
            lines = f.readlines()
            f.close()

        except IOError as e:
            warnings.warn(f"Unable to open input file")
            print("{repr(e)}")
            raise ValueError

        # assumed defaults for simple regions file
        point = "circle"  # only one supported right now
        delta = textoff  # pixels to offset text
        lines = lines[header:]

        text = list()
        x = list()
        y = list()

        for i in range(0, len(lines), 1):
            words = lines[i].split(',')
            x.append(words[0].strip())
            y.append(words[1].strip())
            if(len(words) > 2 and labels):
                text.append(words[2].strip())

        # now write out to a reg file
        out = infile + ".reg"
        f = open(out, 'w')
        for i in range(0, len(lines), 1):
            pline = "image; " + point + \
                "(" + x[i] + "," + y[i] + "," + str(size) + ")\n"
            f.write(pline)
            if(len(text) > 0):
                pline = "image;text(" + str(float(x[i]) + delta) + "," + \
                    str(float(y[i]) + delta) + \
                    "{ " + text[i] + " })# font=\"time 12 bold\"\n"
                f.write(pline)
        f.close()

        self.log.info(f"output reg file saved to: {out}")

    def match(self, coordsys="wcs", frame=True, crop=False, fslice=False,
              scale=False, bin=False, colorbar=False, smooth=False,
              crosshair=False):
        """match all other frames to the current frame.

        Parameters
        ----------

        coordsys: string, optional
            The coordinate system to use

        frame: bool, optional
            Match all other frames to the current frame, using the set coordsys

        crop: bool, optional
            Set the current image display area, using the set coordsys

        fslice: bool, optional
            Match current slice in all frames

        scale: bool, optional
            Match to the current scale for all frames

        bin: bool, optional
            Match to the current binning for all frames

        colorbar: bool, optional
            Match to the current colorbar for all frames

        smooth: bool, optional
            Match to the current smoothing for all frames

        crosshair: bool, optional
            Match the crosshair in all frames, using the current coordsys


        Notes
        -----

        You can only choose one of the options at a time, and the logic will
        select the first True option so set frame=False and something else in
        addition to your choice if you don't want the default option.

        """
        cstring = "match "
        if frame:
            cstring += "frame {0:s}".format(coordsys)
        elif crosshair:
            cstring += "crosshair {0:s}".format(coordsys)
        elif crop:
            cstring += "crop {0:s}".format(coordsys)
        elif fslice:
            cstring += "slice"
        elif bin:
            cstring += "bin"
        elif scale:
            cstring += "scale"
        elif colorbar:
            cstring += "colorbar"
        elif smooth:
            cstring += "smooth"
        self.set(cstring)

    def nancolor(self, color="red"):
        """set the not-a-number color, default is red.

        Parameters
        ----------

        color: string
            The color to use for NAN pixels


        """
        self.set("nan {0:s}".format(color))

    def panto_image(self, x, y):
        """convenience function to change to x,y  physical image coordinates.

        Parameters
        ----------
        x: float
            X location in physical coords to pan to

        y: float
            Y location in physical coords to pan to

        """
        self.set("pan to {0:f} {1:f} image".format(x, y))

    def panto_wcs(self, x, y, system='fk5'):
        """pan to wcs location coordinates in image.

        Parameters
        ----------

        x: string
            The x location to move to, specified using the given system
        y: string
            The y location to move to
        system: string
            The reference system that x and y were specified in, they should
            be understood by DS9

        """
        self.set("pan to {0:s} {1:s} wcs {2:s}".format(x, y, system))

    def rotate(self, value=None, to=False):
        """rotate the current frame (in degrees).

        the current rotation is printed with no params

        Parameters
        ----------

        value: float [degrees]
            Rotate the current frame {value} degrees
            If value is 0, then the current rotation is printed

        to: bool
            Rotate the current frame to the specified value

        """
        if value is None:
            print("Image rotated at {0:s}".format(self.get("rotate")))

        elif to and value >= 0:
            cstring = "rotate to {0:s}".format(str(value))
            self.set(cstring)

        elif value > 0:
            cstring = "rotate {0:s}".format(str(value))
            self.set(cstring)
        cstring = "Image rotated at {0:s}".format(self.get("rotate"))
        self.log.info(cstring)

    def save_regions(self, filename=None):
        """save the regions in the current window to a DS9 style regions file.

        Parameters
        ----------

        filename: string
            The nameof th file to which the regions displayed in the current
            window are saved. If no filename is provided then it will try and
            save the regions to the name of the file in the current display
            with _regions.txt appended

            If a file of that name already exists on disk it will no attempt
            to overwrite it

        """
        regions = self.get("regions save")
        frame = self.frame()

        if frame and not filename:
            filename = self._viewer[frame]['filename'] + "_regions.txt"

        # check if the file already exists
        if not os.access(filename, os.F_OK):
            with open(filename, "w") as region_file:
                region_file.write(regions)
        else:
            warnings.warn(
                "File already exists: {0} try again".format(filename))

    def save_rgb(self, filename=None):
        """save an rgbimage frame as an MEF fits file.

        Parameters
        ----------

        filename: string
            The name of the output fits image

        """
        if not filename:
            print("No filename specified, try again")
        else:
            self.set("save rgbimage {0:s}".format(filename))

    def scale(self, scale='zscale'):
        """The default zscale is the most widely used option.

        Parameters
        ----------
        scale: string
            The scale for ds9 to use, these are set strings of
            [linear|log|pow|sqrt|squared|asinh|sinh|histequ]

        Notes
        -----
        The xpa doesn't return an error if you set an unknown scale,
        it just doesn't do anything, this is true for all the xpa calls
        """
        _help = """Syntax:
           scales available: [linear|log|pow|sqrt|squared|asinh|sinh|histequ]

          [log exp <value>]
          [datasec yes|no]
          [limits <minvalue> <maxvalue>]
          [mode minmax|<value>|zscale|zmax]
          [scope local|global]
          [match]
          [lock [yes|no]]
          [open|close]

          """
        mode_scale = ["zscale", "zmax", "minmax"]
        cstring = f"scale {scale}"

        if scale in mode_scale:
            cstring = f"scale mode {scale}"
        try:
            self.set(cstring)
        except (XpaException, ValueError):
            print(f"{cstring} not valid")
            print(_help)

    def set_region(self, region_string=""):
        """display a region using the specifications in region_string.

        Parameters
        ----------

        region_string: string
            Should take the form of a region string that DS9 is expecting

        Examples
        --------
        set_region("physical ruler 200 300 200 400")
        set_region("line 0 400 3 400 #color=red")

        """
        command = f"regions command {{ {region_string} }}\n"
        self.set(command)

    def showme(self):
        """raise the ds9 window."""
        self.set("raise")

    def showpix(self, close=False):
        """display the pixel value table, close window when done.

        Parameters
        ----------

        close: bool, optional
            If set to True, then the pixel table dialog window is closed


        """
        self.get("pixeltable")
        if close:
            self.set("pixeltable close")

    def snapsave(self, filename=None, format=None, resolution=100):
        """Create a snap shot of the current window, save in specified format.

        This function bypasses the XPA  calling routines to avoid a bug with
        the X11/XPA interface. Instead is uses the os function which
        takes a snapshot of the specified window.

        Parameters
        ----------

        filename: str, optional
           filename of output image, the extension in the filename can also be
           used to specify the format. If no filename is specified, then the
           filename will be constructed from the name of the image displayed
           image with _snap.png appended.

        format: str, optional
           available formats are fits, eps, gif, tiff, jpeg, png
           If no format is specified the filename extension is used

        resolution: int, optional
           1 to 100, for jpeg images

        """
        frame = self.frame()
        if not filename:
            filename = self._viewer[frame]['filename']

        if not format:
            filename += "_snap.png"
        else:
            filename += "_snap." + format

        # screencapture works with OSX
        if sys.platform == "darwin":
            print("Not implemented for darwin platform")
            return None
        else:  # import only works with x11 windows
            cstring = ['import']
            cstring.append('-window ')
            cstring.append(self._xpa_name)
            if "jpeg" in filename:
                cstring.append(' -quality')
                cstring.append(str(resolution))

            cstring.append(filename)

            # self.set(cstring)
            # save the local directory, erase later?
            call(cstring)
            print(f"Image saved to {filename}")
            self.log.info(f"Image saved to {filename}")
            return(filename)

    def grab(self):
        """Make a copy of the image view."""
        valid_fname = self.snapsave(format="png")
        if valid_fname:
            if "nbagg" in get_backend().lower():  # save inside the notebook
                data = mpimage.imread(valid_fname)
                plt.clf()
                return plt.imshow(data, origin="upper")

    def view(self, img):
        """Display numpy image array to current frame.

        Parameters
        ----------
        img: numpy array
            The array containing data, it will be forced to numpy.array()

        """
        frame = self.frame()
        if not frame:
            print("No valid frame")
        else:
            img = np.array(img)
            if img.dtype.type == np.bool8:
                img = img.astype(np.uint8)

            # arrays of 1 through 3 dimensions
            dim = img.ndim
            if dim == 2:
                (ydim, xdim) = img.shape
                dims = "[xdim={0:d},ydim={1:d},".format(xdim, ydim)
            elif dim == 3:
                (zdim, ydim, xdim) = img.shape
                dims = "[xdim={0:d},ydim={1:d},zdim={2:d},".format(xdim,
                                                                   ydim,
                                                                   zdim)
            else:
                raise UnsupportedImageShapeException(repr(img.shape))

            if img.dtype.byteorder in ["=", "|"]:
                dt = img.dtype.newbyteorder(">")
                img = np.array(img, dtype=dt)
                byteorder = ">"
            else:
                byteorder = img.dtype.byteorder

            endianness = {">": ",arch=bigendian",
                          "<": ",arch=littleendian"}[byteorder]

            arr_str = img.tostring()

            try:
                bitpix = self._ImgCode[img.dtype.name]
            except KeyError as e:
                raise UnsupportedDatatypeException(e)

            option = (dims + "bitpix={0:d}{1:s}]".format(bitpix, endianness))
            try:
                self.set("array " + option, arr_str)
                self._set_frameinfo()
                self._viewer[frame]['user_array'] = img
            except XpaException as e:
                raise XpaException(
                    "XPA: {0} : Problem loading array \
                    into frame {1}".format(e, frame))

    def zoomtofit(self):
        """Zoom to fit the image to the viewer."""
        self.zoom("to fit")

    def zoom(self, par="to fit"):
        """Zoom using the specified command.

        Parameters
        ----------
        par: string
            - it can be a number (ranging 0 to 8 effectively), and successive
                    calls continue zooming in the same direction
            - it can be two numbers '4 2', which specify zoom on different axis
            - if can be to a specific value 'to 8' or 'to fit'
            - it can be 'open' to open the dialog box
            - it can be 'close' to close the dialog box (only valid if the box
                    is already open)


        Examples
        --------
        zoom('0.1')

        """
        try:
            self.set("zoom {0:s}".format(str(par)))
        except XpaException:
            print("XPA problem with zoom (probably your zoom "
                  "window is already closed)")

    def show_xpa_commands(self):
        """Print the available XPA commands."""
        print(self.get(''))  # With empty string, all commands are returned

    def reopen(self):
        """Reopen a closed window."""
        print("Not available for DS9, start a new object instead")
        raise NotImplementedError


atexit.register(ds9._purge_tmp_dirs)
atexit.register(ds9._stop_running_process)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""These are default parameters for some of the plotting functions in Imexam
Users can edit this file to set their own defaults before installation,
they could script something that resets the dictionaries themselves,
or we could create a method to let them set from a json or text file maybe
"""

# aperture photometry parameters
aper_phot_pars = {"function": ["aper_phot", ],
                  "center": [True, "Center the object (choose center_type)"],
                  "center_com": [False, "gaussian2d, True=center of mass"],
                  "width": [5, "Width of sky annulus in pixels"],
                  "subsky": [True, "Subtract a sky background?"],
                  "skyrad": [15, "Distance to start sky annulus is pixels"],
                  "radius": [5, "Radius of aperture for star flux"],
                  "zmag": [25., "zeropoint for the magnitude calculation"],
                  "title": [None, "Title of the plot"],
                  "scale": ['zscale', "How to scale the image"],
                  "color_min": [None, "Minimum color value"],
                  "color_max": [None, "Maximum color value"],
                  "cmap": ['Greys', "Matplotlib colormap to use"],
                  "delta": [10, "bounding box for centering measurement"],
                  }

com_center_pars = {"function": ["com_center",],
                   "delta": [10, "bounding box size"],
                   "oversampling": [1., "oversample pixels by"],
                   }

# box statistics
report_stat_pars = {"function": ["report_stat", ],
                    "stat": ["median", "numpy stat name or describe for scipy.stats"],
                    "region_size": [5, "region size in pixels to use"],
                    }

# radial profile plots
radial_profile_pars = {"function": ["radial_profile", ],
                       "title": [None, "Title of the plot"],
                       "xlabel": ["Radius", "The string for the xaxis label"],
                       "ylabel": ["Flux", "The string for the yaxis label"],
                       "pixels": [True, "Plot all pixels at each radius? (False bins the data)"],
                       "fitplot": [True, "Overplot model fit?"],
                       "func": ["Gaussian1D", "Model form to fit"],
                       "center": [True, "Solve for center using 2d Gaussian? [bool]"],
                       "background": [True, "Subtract background? [bool]"],
                       "skyrad": [10., "Background inner radius in pixels, from center of object"],
                       "width": [5., "Background annulus width in pixels"],
                       "magzero": [25., "magnitude zero point"],
                       "rplot": [8., "Plotting radius in pixels"],
                       "pointmode": [True, "plot points instead of lines? [bool]"],
                       "marker": ["o", "The marker character to use, matplotlib style"],
                       "getdata": [False, "print the plotted data values"],
                       "clip": [False, "Sigma clip by np.mean before center fitting?"],
                       "sigma": [4.0, "sigma value for clipping during model fit"],
                       }

# curve of growth
curve_of_growth_pars = {"function": ["curve_of_growth", ],
                        "title": [None, "Title of the plot"],
                        "xlabel": ["radius", "The string for the xaxis label"],
                        "ylabel": ["Encircled Flux", "The string for the yaxis label"],
                        "center": [True, "Fit for center? [bool]"],
                        "background": [True, "Fit and subtract background? [bool]"],
                        "buffer": [25., "Background inner radius in pixels,from center of star"],
                        "width": [5., "Background annulus width in pixels"],
                        "magzero": [25., "magnitude zero point"],
                        "rplot": [8., "Plotting radius in pixels"],
                        "pointmode": [True, "plot points instead of lines? [bool]"],
                        "marker": ["o", "The marker character to use, matplotlib style"],
                        "logx": [False, "log scale x-axis?"],
                        "logy": [False, "log scale y-axis?"],
                        "minflux": [0., "only measure flux above this value"],
                        "getdata": [True, "return the plotted data values"]
                        }

# surface plots
surface_pars = {"function": ["surface", ],
                "title": [None, "Title of the plot"],
                "xlabel": ["X", "The string for the xaxis label"],
                "ylabel": ["Y", "The string for the yaxis label"],
                "zlabel": [None, "Label for zaxis"],
                "ncolumns": [10, "Number of columns"],
                "nlines": [10, "Number of lines"],
                "azim": [None, "azimuthal viewing angle in degrees"],
                "floor": [None, "Minimum value to be contoured"],
                "ceiling": [None, "Maximum value to be contoured"],
                "stride": [1, "step size, higher vals will have less contour"],
                "cmap": ["viridis", "colormap (matplotlib) for display"],
                "fancy": [True, "This aint your grandpas iraf"],
                }

# fit of line in image using model
line_fit_pars = {"function": ["line_fit", ],
                 "func": ["Gaussian1D", "function for fitting (see available)"],
                 "title": [None, "Title of the plot"],
                 "xlabel": ["Column", "The string for the xaxis label"],
                 "ylabel": ["Flux", "The string for the yaxis label"],
                 "background": [False, "Solve for background? [bool]"],
                 "width": [10.0, "Background  width in pixels"],
                 "order": [1, "Polynomial order to fit, 1=line"],
                 "rplot": [15, "Plotting radius in pixels"],
                 "pointmode": [True, "plot points instead of lines? [bool]"],
                 "logx": [False, "log scale x-axis?"],
                 "logy": [False, "log scale y-axis?"],
                 "center": [True, "Recenter around the local max"],
                 "clip": [False, "Sigma clip by the mean before fitting?"],
                 "sigma": [3.0, "sigma value for clipping"],
                 }

# fit of column in image using model
column_fit_pars = {"function": ["column_fit", ],
                   "func": ["Gaussian1D", "function for fitting (see available)"],
                   "title": [None, "Title of the plot"],
                   "xlabel": ["Line", "The string for the xaxis label"],
                   "ylabel": ["Flux", "The string for the yaxis label"],
                   "background": [False, "Solve for background? [bool]"],
                   "width": [10.0, "Background  width in pixels"],
                   "order": [1, "Polynomial order to fit, 1=line"],
                   "rplot": [15., "Plotting radius in pixels"],
                   "pointmode": [True, "plot points instead of lines? [bool]"],
                   "logx": [False, "log scale x-axis?"],
                   "logy": [False, "log scale y-axis?"],
                   "center": [True, "Recenter around the local max"],
                   "clip": [False, "Sigma clip by the mean before fitting?"],
                   "sigma": [3.0, "sigma value for clipping"],
                   }

# contour plots
contour_pars = {"function": ["contour", ],
                "title": [None, "Title of the plot"],
                "xlabel": ["x", "The string for the xaxis label"],
                "ylabel": ["y", "The string for the yaxis label"],
                "ncolumns": [15, "Number of columns"],
                "nlines": [15, "Number of lines"],
                "floor": [None, "Minimum value to be contoured"],
                "ceiling": [None, "Maximum value to be contoured"],
                "ncontours": [8, "Number of contours to be drawn"],
                "linestyles": ["--", "matplotlib linestyle"],
                "label": [True, "Label major contours with their values? [bool]"],
                "cmap": ["viridis", "Colormap (matplotlib style) for image"],
                }

# histogram of values contained in a box
histogram_pars = {"function": ["histogram", ],
                  "title": [None, "Title of the plot"],
                  "xlabel": ["Flux (bin)", "The string for the xaxis label"],
                  "ylabel": ["Count", "The string for the yaxis label"],
                  "ncolumns": [21, "Number of columns"],
                  "nlines": [21, "Number of lines"],
                  "nbins": [100, "Number of bins"],
                  "z1": [None, "Minimum histogram intensity"],
                  "z2": [None, "Maximum histogram intensity"],
                  "pointmode": [True, "plot points instead of lines? [bool]"],
                  "marker": ["o", "The marker character to use, matplotlib style"],
                  "logx": [False, "log scale x-axis?"],
                  "logy": [False, "log scale y-axis?"],
                  }


# plots the values in a line
lineplot_pars = {"function": ["line_plot", ],
                 "title": [None, "Title of the plot"],
                 "xlabel": ["Column", "The string for the xaxis label"],
                 "ylabel": ["Pixel Value", "The string for the yaxis label"],
                 "pointmode": [False, "plot points instead of lines? [bool]"],
                 "marker": ["o", "The marker character to use, matplotlib style"],
                 "logx": [False, "log scale x-axis?"],
                 "logy": [False, "log scale y-axis?"],
                 "xmin": [0, "xaxis min value"],
                 "xmax": [None, "xaxis max value"],
                 "ymin": [0, "yaxis min value"],
                 "ymax": [None, "yaxis max value"],
                 }
# plot the values in a column
colplot_pars = {"function": ["column_plot", ],
                "title": [None, "Title of the plot"],
                "xlabel": ["Line", "The string for the xaxis label"],
                "ylabel": ["Pixel Value", "The string for the yaxis label"],
                "pointmode": [False, "plot points instead of lines? [bool]"],
                "marker": ["o", "The marker character to use, matplotlib style"],
                "logx": [False, "log scale x-axis?"],
                "logy": [False, "log scale y-axis?"],
                "xmin": [0, "xaxis min value"],
                "xmax": [None, "xaxis max value"],
                "ymin": [0, "yaxis min value"],
                "ymax": [None, "yaxis max value"],
                }

# cut out a smaller image with image coordinates
cutout_pars = {"function": ["cutout", ],
               "size": [20, "size of the image to cutout"],
               }



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Functions to help enable math and fitting in the main examine code"""

import numpy as np
import warnings

from astropy.modeling import models, fitting
from astropy.stats import gaussian_sigma_to_fwhm, sigma_clip


def gfwhm(sigmax, sigmay=None):
    """Compute the Gaussian full width half max.

    Parameters
    ----------

    sigmax: float
        The input sigma to use

    sigmay: float
        The input sigma to use

    Returns
    -------
    The FWHM tuple for the Gaussian

    """
    if sigmax is None:
        print("Need at least one sigma value for Gaussian FWHM")
        return (None, None)

    fwhmx = gaussian_sigma_to_fwhm * sigmax
    if sigmay is None:
        fwhmy = fwhmx
    else:
        fwhmy = gaussian_sigma_to_fwhm * sigmay

    return (fwhmx, fwhmy)


def mfwhm(alpha=0, gamma=0):
    """Compute the Moffat full width half max.

    Parameters
    ----------

    sigmax: float
        The input sigma to use

    sigmay: float
        The input sigma to use

    Returns
    -------
    The FWHM tuple for a Moffat function
    fwhm = 2* alpha * sqrt(2^(1/gamma) -1 ))
    """
    if alpha == 0 or gamma == 0:
        print("Need alpha AND gamma values for moffat FWHM")
        return None

    return 2 * alpha * np.sqrt(2 ** (1 / gamma) - 1)


def fit_moffat_1d(data, gamma=2., alpha=1., sigma_factor=0.,
                  center_at=None, weighted=False):
    """Fit a 1D moffat profile to the data and return the fit.

    Parameters
    ----------
    data: 2D data array
        The input sigma to use
    gamma: float (optional)
        The input gamma to use
    alpha: float (optional)
        The input alpha to use
    sigma_factor: float (optional)
        If sigma>0 then sigma clipping of the data is performed
        at that level
    center_at: float or None
        None by default, set to value to use as center
    weighted: bool
        if weighted is True, then weight the values by basic
        uncertainty hueristic

    Returns
    -------
    The fitted 1D moffat model for the data

    """
    # data is assumed to already be chunked to a reasonable size
    ldata = len(data)

    # guesstimate mean
    if center_at:
        x0 = 0
    else:
        x0 = int(ldata / 2.)

    # assumes negligable background
    if weighted:
        z = np.nan_to_num(1. / np.sqrt(data))  # use as weight

    x = np.arange(ldata)

    # Initialize the fitter
    fitter = fitting.LevMarLSQFitter()
    if sigma_factor > 0:
        fit = fitting.FittingWithOutlierRemoval(fitter,
                                                sigma_clip,
                                                niter=3,
                                                sigma=sigma_factor)
    else:
        fit = fitter

    # Moffat1D + constant
    model = (models.Moffat1D(amplitude=max(data),
                             x_0=x0,
                             gamma=gamma,
                             alpha=alpha) +
             models.Polynomial1D(c0=data.min(), degree=0))

    with warnings.catch_warnings():
        # Ignore model linearity warning from the fitter
        warnings.simplefilter('ignore')
        if weighted:
            results = fit(model, x, data, weights=z)
        else:
            results = fit(model, x, data)

    # previous yield amp, ycenter, xcenter, sigma, offset
    # if sigma clipping is used, results is a tuple of data, model
    if sigma_factor > 0:
        return results[1]
    else:
        return results


def fit_gauss_1d(radius, flux, sigma_factor=0, center_at=None, weighted=False):
    """Fit a 1D gaussian to the data and return the fit.

    Parameters
    ----------
    radius: array of float
        set center_at and the center will be taken there
    flux: array of float
        values should correspond to radius array
    sigma_factor: float (optional)
        If sigma>0 then sigma clipping of the data is performed
        at that level
    center_at: float or None
        None by default, set to value to use as center
        If the value is None, center will be set at half the
        array size.
    weighted: bool
        if weighted is True, then weight the values by basic
        uncertainty hueristic

    Returns
    -------
    The fitted 1D gaussian model for the data.
    """

    if radius.shape != flux.shape:
        raise ValueError("Expected same sizes for radius and flux")

    # guesstimate the mean
    # assumes ordered radius
    if center_at is None:
        delta = int(len(radius) / 2.)
    else:
        delta = center_at

    # assumes negligable background
    if weighted:
        z = np.nan_to_num(np.log(flux))

    # Initialize the fitter
    fitter = fitting.LevMarLSQFitter()
    if sigma_factor > 0:
        fit = fitting.FittingWithOutlierRemoval(fitter,
                                                sigma_clip,
                                                niter=3,
                                                sigma=sigma_factor)
    else:
        fit = fitter

    # Gaussian1D + a constant
    model = (models.Gaussian1D(amplitude=flux.max() - flux.min(),
                               mean=delta, stddev=1.) +
             models.Polynomial1D(c0=flux.min(), degree=0))

    with warnings.catch_warnings():
        # Ignore model linearity warning from the fitter
        warnings.simplefilter('ignore')
        if weighted:
            results = fit(model, radius, flux, weights=z)
        else:
            results = fit(model, radius, flux)

    # previous yield amp, ycenter, xcenter, sigma, offset
    # if sigma clipping is used, results is a tuple of data, model
    if sigma_factor > 0:
        return results[1]
    else:
        return results


def fit_gaussian_2d(data, sigma=3., theta=0., sigma_factor=0):
    """center the data  by fitting a 2d gaussian to the region.

    Parameters
    ----------

    data: 2D float array
        should be a 2d array, the initial center is used to estimate
        the fit center
    sigma: float (optional)
        The sigma value for the starting gaussian model
    theta: float(optional)
        The theta value for the starting gaussian model
    sigma_factor: float (optional)
        If sigma_factor > 0 then clipping will be performed
        on the data during the model fit

    Returns
    -------
    The full gaussian fit model, from which the center can be extracted

    """
    # use a smaller bounding box so that we are only fitting the local data
    delta = int(len(data) / 2)  # guess the center
    amp = data.max() - data.min()  # guess the amplitude
    ldata = len(data)
    yy, xx = np.mgrid[:ldata, :ldata]

    # Initialize the fitter
    fitter = fitting.LevMarLSQFitter()
    if sigma_factor > 0:
        fit = fitting.FittingWithOutlierRemoval(fitter,
                                                sigma_clip,
                                                niter=3,
                                                sigma=sigma_factor)
    else:
        fit = fitter

    # Gaussian2D(amp,xmean,ymean,xstd,ystd,theta) + a constant
    model = (models.Gaussian2D(amp, delta, delta, sigma, sigma, theta) +
             models.Polynomial2D(c0_0=data.min(), degree=0))
    with warnings.catch_warnings():
        # Ignore model linearity warning from the fitter
        warnings.simplefilter('ignore')
        results = fit(model, xx, yy, data)

    # previous yield amp, ycenter, xcenter, sigma, offset
    # if sigma clipping is used, results is a tuple of data, model
    if sigma_factor > 0:
        return results[1]
    else:
        return results


def fit_mex_hat_1d(data, sigma_factor=0, center_at=None, weighted=False):
    """Fit a 1D Mexican Hat function to the data.

    Parameters
    ----------

    data: 2D float array
        should be a 2d array, the initial center is used to estimate
        the fit center
    sigma_factor: float (optional)
        If sigma_factor > 0 then clipping will be performed
        on the data during the model fit
    center_at: float or None
        None by default, set to value to use as center
    weighted: bool
        if weighted is True, then weight the values by basic
        uncertainty hueristic

    Returns
    -------
    The the fit model for mexican hat 1D function
    """
    ldata = len(data)
    if center_at:
        x0 = 0
    else:
        x0 = int(ldata / 2.)

    # assumes negligable background
    if weighted:
        z = np.nan_to_num(1. / np.sqrt(data))  # use as weight

    x = np.arange(ldata)
    fixed_pars = {"x_0": True}

    # Initialize the fitter
    fitter = fitting.LevMarLSQFitter()
    if sigma_factor > 0:
        fit = fitting.FittingWithOutlierRemoval(fitter,
                                                sigma_clip,
                                                niter=3,
                                                sigma=sigma_factor)
    else:
        fit = fitter

    # Mexican Hat 1D + constant
    model = (models.MexicanHat1D(amplitude=np.max(data),
                                 x_0=x0,
                                 sigma=2.,
                                 fixed=fixed_pars) +
             models.Polynomial1D(c0=data.min(), degree=0))

    with warnings.catch_warnings():
        # Ignore model linearity warning from the fitter
        warnings.simplefilter('ignore')
        if weighted:
            results = fit(model, x, data, weights=z)
        else:
            results = fit(model, x, data)

    # previous yield amp, ycenter, xcenter, sigma, offset
    if sigma_factor > 0:
        return results[1]
    else:
        return results


def fit_airy_2d(data, x=None, y=None, sigma_factor=0):
    """Fit an AiryDisk2D model to the data.

    Parameters
    ----------

    data: 2D float array
        should be a 2d array, the initial center is used to estimate
        the fit center
    x: float (optional)
        xcenter location
    y: float (optional)
        ycenter location
    sigma_factor: float (optional)
        If sigma_factor > 0 then clipping will be performed
        on the data during the model fit

    Returns
    -------
    The the fit model for Airy2D function

    """
    delta = int(len(data) / 2)  # guess the center
    ldata = len(data)

    if x is None:
        x = delta
    if y is None:
        y = delta
    fixed_pars = {"x_0": True, "y_0": True}  # hold these constant
    yy, xx = np.mgrid[:ldata, :ldata]

    # Initialize the fitter
    fitter = fitting.LevMarLSQFitter()
    if sigma_factor > 0:
        fit = fitting.FittingWithOutlierRemoval(fitter,
                                                sigma_clip,
                                                niter=3,
                                                sigma=sigma_factor)
    else:
        fit = fitter

    # AiryDisk2D(amplitude, x_0, y_0, radius) + constant
    model = (models.AiryDisk2D(np.max(data),
                               x_0=x,
                               y_0=y,
                               radius=delta,
                               fixed=fixed_pars) +
             models.Polynomial2D(c0_0=data.min(), degree=0))
    with warnings.catch_warnings():
            # Ignore model warnings for new_plot_window
            warnings.simplefilter('ignore')
            results = fit(model, xx, yy, data)

    if sigma_factor > 0:
        return results[1]
    else:
        return results


def fit_poly_n(data, deg=1, sigma_factor=0):
    """Fit a Polynomial 1D model to the data.

    Parameters
    ----------

    data: float array
        should be a 1d or 2d array
    deg: int
        The degree of polynomial to fit
    sigma_factor: float (optional)
        If sigma_factor > 0 then clipping will be performed
        on the data during the model fit

    Returns
    -------
    The the polynomial fit model for the function
    """
    if len(data) < deg + 1:
        raise ValueError("fit_poly_n: Need more data for fit")

    # define the model
    poly = models.Polynomial1D(deg)

    # set the axis range for fitting
    ax = np.arange(len(data))

    # define the fitter
    fitter = fitting.LinearLSQFitter()

    if sigma_factor > 0:
        fit = fitting.FittingWithOutlierRemoval(fitter,
                                                sigma_clip,
                                                sigma=sigma_factor,
                                                niter=3)
    else:
        fit = fitter

    try:
        result = fit(poly, ax, data)
    except ValueError:
        result = None

    if sigma_factor > 0:
        result = result[1]

    return result



"""Licensed under a 3-clause BSD style license - see LICENSE.rst.

   This class implements IRAF/imexamine type capabilities
   for providing powerful diagnostic quick-look tools.

   However, the power of this python tool is that it is essentially a library
   of plotting and analysis routines which can be directed towards
   any viewer. It can also be used without connecting to any viewer
   since the calls take only data,x,y information. This means that
   given a data array and a list of x,y positions you can creates
   plots without havin to interact with the viewers.

   Users can also register a custom function with the class
   and have it available for use in either case.

   The plots which are made are fully customizable

 """

import matplotlib.pyplot as plt
# turn on interactive mode for plotting
# so that plotting becomes non-blocking
# if not plt.isinteractive():
#     plt.ion()

import warnings
import numpy as np
import sys
import logging
import tempfile
from copy import deepcopy


from matplotlib import get_backend

from astropy.io import fits
from astropy.modeling import models
from astropy.visualization import ZScaleInterval

try:
    from scipy import stats
    scipy_installed = True
except ImportError:
    scipy_installed = False
    print("Scipy not installed, describe stat unavailable")

from . import math_helper
from . import imexam_defpars
from .util import set_logging

if sys.version_info.major < 3:
    PY3 = False
else:
    PY3 = True

# enable display plot in iPython notebook
try:
    from io import StringIO
except ImportError:
    from cString import StringIO

try:
    import photutils
    photutils_installed = True
    from photutils.centroids import centroid_com
    # account for API change
    from packaging import version
    photutils_version = version.parse(photutils.__version__)
except ImportError:
    print("photutils not installed, photometry functionality "
          "in imexam() not available")
    photutils_installed = False

__all__ = ["Imexamine"]


class Imexamine:
    """The imexamine class controls plotting and analysis functions."""

    def __init__(self):
        """do imexamine like routines on the current frame.

        read the returned cursor key value to decide what to do

        region_size is the default radius or side of the square for stat info
        """
        self.set_option_funcs()  # define the dictionary of keys and functions
        self._data = np.zeros(0)  # the data array
        self._datafile = ""  # the file from which the data came
        # read from imexam_defpars which contains dicts
        self._define_default_pars()
        # default plot name saved with "s" key
        self.plot_name = "imexam_plot.pdf"
        # let users have multiple plot windows, the list stores their names
        self._plot_windows = list()
        # this contains the name of the current plotting window
        self._figure_name = "imexam"
        self._plot_windows.append(self._figure_name)
        self._reserved_keys = ['q', '2']  # not to be changed with user funcs
        self._fit_models = ["Gaussian1D",
                            "Moffat1D",
                            "MexicanHat1D",
                            "AiryDisk2D",
                            "Polynomial1D"]

        # see if the package logger was already started
        self.log = logging.getLogger(__name__)
        self.log = set_logging()

        # save the backend that is in use for plotting reference
        self._mpl_backend = get_backend().lower()

    def setlog(self, filename=None, on=True, level=logging.INFO):
            """Turn on and off logging to a logfile or the screen.

            Parameters
            ----------
            filename: str, optional
                Name of the  output file to record log information
            on: bool, optional
                True by default, turn the logging on or off
            level: logging class, optional
                set the level for logging messages, turn off screen messages
                by setting to logging.CRITICAL

            """
            self.log = set_logging(filename, on, level)

    def _close_plots(self):
        """Make sure to release plot memory at end of exam loop."""
        for plot in self._plot_windows:
            plt.close()

    def close(self):
        """For use with the Imexamine object standalone."""
        self._close_plots()
        self.log = set_logging(on=False)

    def show_fit_models(self):
        """Print the available astropy models for plot fits."""
        self.log.info(f"The available astropy models for fitting"
                      f"are: {self._fit_models}")

    def set_option_funcs(self):
        """Define the dictionary which maps imexam keys to their functions.

         Notes
         -----
         The user can modify this dictionary to add or change options,
         the first item in the tuple is the associated function
         the second item in the tuple is the description of what the function
         does when that key is pressed
        """
        self.imexam_option_funcs = {'a': (self.aper_phot, 'Aperture sum, with radius region_size '),
                                    'j': (self.line_fit, '1D [Gaussian1D default] line fit '),
                                    'k': (self.column_fit, '1D [Gaussian1D default] column fit'),
                                    'm': (self.report_stat, 'Square region stats, in [region_size],default is median'),
                                    'x': (self.show_xy_coords, 'Return x,y,value of pixel'),
                                    'y': (self.show_xy_coords, 'Return x,y,value of pixel'),
                                    'l': (self.plot_line, 'Return line plot'),
                                    'c': (self.plot_column, 'Return column plot'),
                                    'g': (self.curve_of_growth, 'Return curve of growth plot'),
                                    'r': (self.radial_profile, 'Return the radial profile plot'),
                                    'h': (self.histogram, 'Return a histogram in the region around the cursor'),
                                    'e': (self.contour, 'Return a contour plot in a region around the cursor'),
                                    's': (self.save_figure, 'Save current figure to disk as [plot_name]'),
                                    'b': (self.gauss_center, 'Return the 2D gauss fit center of the object'),
                                    'd': (self.com_center, 'Return the Center of Mass fit center of the object'),
                                    'w': (self.surface, 'Display a surface plot around the cursor location'),
                                    '2': (self.new_plot_window, 'Make the next plot in a new window'),
                                    't': (self.cutout, 'Make a fits image cutout using pointer location')
                                    }

    def print_options(self):
        """Print the imexam options to screen."""
        keys = self.get_options()
        for key in keys:
            print(f"{key} {self.option_descrip(key)}")

    def do_option(self, x, y, key):
        """Run the imexam option.

        Parameters
        ----------

        x: int
            The x location of the cursor or data point

        y: int
            The y location of the cursor or data point

        key: string
            The key which was pressed

        """
        self.log.debug(f"pressed: {key}, {self.imexam_option_funcs[key][0].__name__}")
        # dont require input for saving the active figure
        if key == 's':
            self.imexam_option_funcs[key][0]()
        else:
            self.imexam_option_funcs[key][0](x, y, self._data)

    def get_options(self):
        """Return the imexam options as a key list."""
        keys = sorted(self.imexam_option_funcs.keys())
        return keys

    def option_descrip(self, key, field=1):
        """Return the looked up dictionary of options.

        Parameters
        ----------
        key: string
            The key which was pressed, it relates to the function to call

        field: int
            This tells where in the option dictionary the function name
            can be found
        """
        return self.imexam_option_funcs[key][field]

    def set_data(self, data=np.zeros(0)):
        """initialize the data that imexamine uses."""
        self._data = data

    def set_plot_name(self, filename=None):
        """set the default plot name for the "s" key.

        Parameters
        ----------
        filename: string
          The name which is used to save the current plotting window to a file
          The extension on the name decides which file type is used
        """
        if filename is None:
            warnings.warn("No filename provided")
        else:
            self.plot_name = filename

    def get_plot_name(self):
        """return the default plot name."""
        return self.plot_name

    def _define_default_pars(self):
        """Set all pars to their defaults, stored in a file with dicts."""
        self.aper_phot_def_pars = imexam_defpars.aper_phot_pars
        self.radial_profile_def_pars = imexam_defpars.radial_profile_pars
        self.curve_of_growth_def_pars = imexam_defpars.curve_of_growth_pars
        self.surface_def_pars = imexam_defpars.surface_pars
        self.line_fit_def_pars = imexam_defpars.line_fit_pars
        self.column_fit_def_pars = imexam_defpars.column_fit_pars
        self.contour_def_pars = imexam_defpars.contour_pars
        self.histogram_def_pars = imexam_defpars.histogram_pars
        self.lineplot_def_pars = imexam_defpars.lineplot_pars
        self.colplot_def_pars = imexam_defpars.colplot_pars
        self.histogram_def_pars = imexam_defpars.histogram_pars
        self.contour_def_pars = imexam_defpars.contour_pars
        self.report_stat_def_pars = imexam_defpars.report_stat_pars
        self.cutout_def_pars = imexam_defpars.cutout_pars
        self.com_center_def_pars = imexam_defpars.com_center_pars
        self._define_local_pars()

    def _define_local_pars(self):
        """Set a copy of the default pars that users can alter."""
        self.aper_phot_pars = deepcopy(self.aper_phot_def_pars)
        self.radial_profile_pars = deepcopy(self.radial_profile_def_pars)
        self.curve_of_growth_pars = deepcopy(self.curve_of_growth_def_pars)
        self.surface_pars = deepcopy(self.surface_def_pars)
        self.line_fit_pars = deepcopy(self.line_fit_def_pars)
        self.column_fit_pars = deepcopy(self.column_fit_def_pars)
        self.contour_pars = deepcopy(self.contour_def_pars)
        self.histogram_pars = deepcopy(self.histogram_def_pars)
        self.lineplot_pars = deepcopy(self.lineplot_def_pars)
        self.colplot_pars = deepcopy(self.colplot_def_pars)
        self.histogram_pars = deepcopy(self.histogram_def_pars)
        self.contour_pars = deepcopy(self.contour_def_pars)
        self.report_stat_pars = deepcopy(self.report_stat_def_pars)
        self.cutout_pars = deepcopy(self.cutout_def_pars)
        self.com_center_pars = deepcopy(self.com_center_def_pars)

    def unlearn_all(self):
        """reset the default parameters for all functions."""
        self._define_local_pars()

    def new_plot_window(self, x, y, data=None):
        """make the next plot in a new plot window.

        Notes
        -----
        x,y, data, are not used here, but the calls are setup to take them
        for all imexam options. Is there a better way to do the calls in
        general? Once the new plotting window is open all plots will be
        directed towards it. The old window cannot be used again.

        """
        if data is None:
            data = self._data
        self._figure_name = "imexam" + str(len(self._plot_windows) + 1)
        self._plot_windows.append(self._figure_name)
        self.log.info(f"Plots now directed towards {self._figure_name}")

    def plot_line(self, x, y, data=None, fig=None):
        """line plot of data at point x.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        fig: figure object for redirect
            Used for interaction with the ginga GUI
        """
        if data is None:
            data = self._data
        x = int(x)
        y = int(y)
        self.log.info(f"Line at {x} {y}")

        pfig = fig
        if fig is None:
            fig = plt.figure(self._figure_name)
        fig.clf()
        fig.add_subplot(111)
        ax = fig.gca()

        if self.lineplot_pars["title"][0] is None:
            ax.set_title(f"{self._datafile} line at {y}")
        ax.set_xlabel(self.lineplot_pars["xlabel"][0])
        ax.set_ylabel(self.lineplot_pars["ylabel"][0])

        if not self.lineplot_pars["xmax"][0]:
            xmax = len(data[y, :])
        else:
            xmax = self.lineplot_pars["xmax"][0]
        ax.set_xlim(self.lineplot_pars["xmin"][0], xmax)

        if self.lineplot_pars["logx"][0]:
            ax.set_xscale("log")
        if self.lineplot_pars["logy"][0]:
            ax.set_yscale("log")

        if bool(self.lineplot_pars["pointmode"][0]):
            ax.plot(data[y, :], self.lineplot_pars["marker"][0])
        else:
            ax.plot(data[y, :])

        if pfig is None and 'nbagg' not in self._mpl_backend:
            plt.draw()
            plt.pause(0.001)
        else:
            fig.canvas.draw_idle()

    def plot_column(self, x, y, data=None, fig=None):
        """column plot of data at point y.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        fig: figure name for redirect
            Used for interaction with the ginga GUI
        """
        if data is None:
            data = self._data
        x = int(x)
        y = int(y)
        self.log.info(f"Column at {x} {y}")

        pfig = fig
        if fig is None:
            fig = plt.figure(self._figure_name)
        fig.clf()
        fig.add_subplot(111)
        ax = fig.gca()

        if self.colplot_pars["title"][0] is None:
            ax.set_title(f"{self._datafile} column at {x}")
        else:
            ax.set_title(self.colplot_pars["title"][0])
        ax.set_xlabel(self.colplot_pars["xlabel"][0])
        ax.set_ylabel(self.colplot_pars["ylabel"][0])

        if not self.colplot_pars["xmax"][0]:
            xmax = len(data[:, x])
        else:
            xmax = self.colplot_pars["xmax"][0]
        ax.set_xlim(self.colplot_pars["xmin"][0], xmax)

        if self.colplot_pars["logx"][0]:
            ax.set_xscale("log")

        if self.colplot_pars["logy"][0]:
            ax.set_yscale("log")
        if bool(self.colplot_pars["pointmode"][0]):
            ax.plot(data[:, x], self.colplot_pars["marker"][0])
        else:
            ax.plot(data[:, x])

        if pfig is None and 'nbagg' not in self._mpl_backend:
            plt.draw()
            plt.pause(0.001)
        else:
            fig.canvas.draw_idle()

    def show_xy_coords(self, x, y, data=None):
        """print the x,y,value to the screen.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        """
        if data is None:
            data = self._data
        info = f"{x} {y}  {data[int(y), int(x)]}"
        self.log.info(info)

    def report_stat(self, x, y, data=None):
        """report the statisic of values in a box with side region_size.

        The statistic can be any numpy function

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        """
        if data is None:
            data = self._data

        region_size = self.report_stat_pars["region_size"][0]
        name = self.report_stat_pars["stat"][0]
        dist = region_size / 2
        xmin = int(x - dist)
        xmax = int(x + dist)
        ymin = int(y - dist)
        ymax = int(y + dist)

        if (("describe" in name) and (scipy_installed)):
            try:
                stat = getattr(stats, "describe")
                nobs, minmax, mean, var, skew, kurt = stat(data[ymin:ymax,
                                                           xmin:xmax].flatten())
                pstr = (f"[{ymin}:{ymax},{xmin}:{xmax}] {name}: \nnobs: "
                        f"{nobs}\nminmax: {minmax}\nmean {mean}\nvariance: {var}\nskew: "
                        f"{skew}\nkurtosis: {kurt}")
            except AttributeError:
                warnings.warn("Invalid stat specified")
        else:
            try:
                stat = getattr(np, name)
                pstr = f"[{ymin}:{ymax},{xmin}:{xmax}] {name}: {stat(data[ymin:ymax, xmin:xmax])}"
            except AttributeError:
                warnings.warn("Invalid stat specified")

        self.log.info(pstr)

    def save_figure(self, fig=None):
        """Save to file the figure that's currently displayed.

        this is used for the imexam loop, because there is a standard api
        for the loop

        Parameters
        ----------
        data: numpy array
            The data array to work on
        fig: figure for redirect
            Used for interaction with the ginga GUI
        """
        if fig is None:
            fig = plt.figure(self._figure_name)
        ax = fig.gca()
        fig.savefig(self.plot_name)
        pstr = f"plot saved to {self.plot_name}"
        self.log.info(pstr)


    def save(self, filename=None, fig=None):
        """Save to file the figure that's currently displayed.

        this is used for the standalone plotting

        Parameters
        ----------
        filename: string
            Name of the file the plot will be saved to. The extension
            on the filename determines the filetype
        fig: figure name for redirect
            Used for interaction with the ginga GUI
        """
        if filename:
            self.set_plot_name(filename)
        else:
            self.set_plot_name(self._figure_name + ".pdf")

        if fig is None:
            fig = plt.figure(self._figure_name)
        ax = fig.gca()
        fig.savefig(self.plot_name)
        pstr = f"plot saved to {self.plot_name}"
        self.log.info(pstr)

    def aper_phot(self, x, y, data=None,
                  genplot=True, fig=None,
                  error=None):
        """Perform aperture photometry.

        Uses photutils functions, photutils must be available

        Parameters
        ----------
        x: int
            The x location of the object

        y: int
            The y location of the object

        data: numpy array
            The data array to work on

        genplot: bool
            plot the apertures to a figure; if false then the
            tuple of (apertures, annulus_apertures,rawflux_table, sky_per_pix)
            is returned.

        fig: figure object for redirect
            Used for interaction with the ginga GUI

        error: float array
            If error is not None, then it should be given the
            error array for the corresponding data image.
            error is assumed to include all sources of error,
            including the Poisson error of the sources
            See the docs for photutils for more details.
            The returned table will include a 'aperture_sum_err' column
            in addition to 'aperture_sum'. 'aperture_sum_err'
            provides the propagated uncertainty associated with
            'aperture_sum'.


        Returns
        -------
        plot or the tuple of apertures, annulus_apertures, rawflux_table, sky_per_pix.
        Where apertures and annulus_apertures are photuils objects, or None

        """
        if data is None:
            data = self._data
        if error is not None:
            if (data.shape != error.shape):
                raise AttributeError("Data and error arrays don't match")

        center = False
        if not photutils_installed:
            self.log.warning("Install photutils to enable")
        else:
            if self.aper_phot_pars["center"][0]:
                center = True
                delta = int(self.aper_phot_pars["delta"][0])
                if self.aper_phot_pars["center_com"][0]:
                    # get center from center of mass
                    xx, yy = self.com_center(x, y, data=data, delta=delta)
                    sigma = 0.
                    sigmay = 0.
                else:
                    amp, xx, yy, sigma, sigmay = self.gauss_center(x, y, data,
                                                                 delta=delta)

            radius = self.aper_phot_pars["radius"][0]
            width = int(self.aper_phot_pars["width"][0])
            inner = int(self.aper_phot_pars["skyrad"][0])
            subsky = bool(self.aper_phot_pars["subsky"][0])

            outer = inner + width

            apertures = photutils.CircularAperture((xx, yy), radius)
            rawflux_table = photutils.aperture_photometry(
                data,
                apertures,
                subpixels=1,
                error=error,
                method="center")

            sky_per_pix = 0.
            annulus_apertures = None
            if subsky:
                annulus_apertures = photutils.CircularAnnulus(
                    (xx, yy), r_in=inner, r_out=outer)
                bkgflux_table = photutils.aperture_photometry(
                    data,
                    annulus_apertures)

                # to calculate the mean local background, divide the circular
                # annulus aperture sums by the area fo the circular annulus.
                # The bkg sum with the circular aperture is then
                # then mean local background tims the circular apreture area.
                if photutils_version >= version.parse('0.7'):
                    aperture_area = apertures.area
                    annulus_area = annulus_apertures.area
                else:
                    aperture_area = apertures.area()
                    annulus_area = annulus_apertures.area()

                bkg_sum = float(
                    (bkgflux_table['aperture_sum'] *
                     aperture_area /
                     annulus_area)[0])
                total_flux = rawflux_table['aperture_sum'][0] - bkg_sum
                sky_per_pix = float(bkgflux_table['aperture_sum'] /
                                    annulus_area)

            else:
                total_flux = float(rawflux_table['aperture_sum'][0])

            # compute the magnitude of the sky corrected flux
            magzero = float(self.aper_phot_pars["zmag"][0])
            mag = magzero - 2.5 * (np.log10(total_flux))

            # Construct the output strings (header and parameter values)
            pheader = f"x\ty\tradius\tflux\tmag(zpt={magzero:0.2})\t"
            pstr = f"\n{x:.2f}\t{y:0.2f}\t{radius}\t{total_flux:0.2}\t{mag:0.2}\t"

            if subsky:
                pheader += "sky/pix\t"
                pstr += f"{sky_per_pix:0.3f}\t"
            if center:
                # center of mass estimator
                if self.aper_phot_pars["center_com"][0]:
                    pheader += "center of mass(x,y)"
                    pstr += f"{xx:0.2f},{yy:0.2f}"
                else:
                    pheader += "fwhm(x,y)"
                    fwhmx, fwhmy = math_helper.gfwhm(sigma, sigmay)
                    pstr += f"{sigma:0.2f},{sigmay:0.2f}"

            pheader = pheader.expandtabs(15)
            pstr = pstr.expandtabs(15)

            self.log.info(pheader + pstr)
            if genplot:
                pfig = fig
                if fig is None:
                    # Make sure figure is square so round stars look round
                    fig = plt.figure(self._figure_name, figsize=[5, 5])
                fig.clf()
                fig.add_subplot(111)
                ax = fig.gca()

                if self.aper_phot_pars["title"][0] is None:
                    title = f"x= {xx:0.2f}, y={yy:0.2f}, flux={total_flux:0.1f}, \nmag={mag:0.1f}, sky={sky_per_pix:0.1f}"
                    if center:
                        if self.aper_phot_pars["center_com"][0]:
                            title+= f", CoM({xx:0.2f},{yy:0.2f})"
                        else:
                            title += f", FWHM={math_helper.gfwhm(sigma)[0]:0.2f}"
                    ax.set_title(title)
                else:
                    ax.set_title(self.aper_phot_pars["title"][0])

                if self.aper_phot_pars['scale'][0] == 'zscale':
                    zs = ZScaleInterval()
                    color_range = zs.get_limits(data)
                else:
                    color_range = [self.aper_phot_pars['color_min'][0],
                                   self.aper_phot_pars['color_max'][0]]

                pad = outer * 1.2  # XXX TODO: Bad magic number
                print(xx,yy,pad)
                ax.imshow(data[int(yy - pad):int(yy + pad),
                               int(xx - pad):int(xx + pad)],
                          vmin=color_range[0], vmax=color_range[1],
                          extent=[int(xx - pad), int(xx + pad),
                                  int(yy - pad), int(yy + pad)], origin='lower',
                          cmap=self.aper_phot_pars['cmap'][0])

                apertures.plot(ax=ax, color='green', alpha=0.75, lw=3)
                if subsky:
                    annulus_apertures.plot(ax=ax, color='red', alpha=0.75, lw=3)

                if pfig is None:
                    plt.draw()
                    plt.pause(0.001)
                else:
                    fig.canvas.draw_idle()
            else:
                return (apertures, annulus_apertures, rawflux_table, sky_per_pix)

    def line_fit(self, x, y, data=None, form=None, genplot=True, fig=None, col=False):
        """compute the 1D fit to the line of data using the specified form.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        form: string
            This is the functional form specified in the line fit parameters
            see show_fit_models()
        genplot: bool
            produce the plot or return the fit
        fig: figure for redirect
            Used for interaction with the ginga GUI
        col: bool (False)
            Plot column instead of line

        Notes
        -----
        The background is currently ignored

        If centering is True in the parameter set, then the center
        is fit with a 2d gaussian, not performed for Polynomial1D
        """

        # Set which parameters to use
        if col:
            pars = self.column_fit_pars
        else:
            pars = self.line_fit_pars

        if data is None:
            data = self._data

        if form is None:
            fitform = getattr(models, pars["func"][0])
        else:
            if form in self._fit_models:
                fitform = getattr(models, form)
            else:
                raise ValueError(f"Functional form not in available: {self._fit_models}")
        self.log.info(f"using model: {fitform}")

        # Used for Polynomial1D fitting
        degree = int(pars["order"][0])

        delta = int(pars["rplot"][0])
        if delta >= len(data) / 4:  # help with small data arrays and defaults
            delta = delta / 2
        delta = int(delta)

        xx = int(x)
        yy = int(y)

        # fit the center with a 2d gaussian
        if pars["center"][0]:
            if fitform.name != "Polynomial1D":
                amp, xout, yout, sigma, sigmay = self.gauss_center(xx,
                                                                   yy,
                                                                   data,
                                                                   delta=delta)
                if (xout < 0 or yout < 0 or xout > data.shape[1] or
                   yout > data.shape[0]):
                        self.log.warning("Problem with centering, "
                                         "pixel coords")
                else:
                    xx = int(xout)
                    yy = int(yout)
        if col:
            line = data[:, xx]
            chunk = line[yy - delta:yy + delta]
            delta_add = yy - delta
        else:
            line = data[yy, :]
            chunk = line[xx - delta: xx + delta]
            delta_add = xx - delta

        # This factor is passed to the fitter
        if pars["clip"][0]:
            sig_factor = pars["sigma"][0]
        else:
            sig_factor = 0

        # fit model to data
        if fitform.name == "Gaussian1D":
            xr = np.arange(len(chunk))
            fitted = math_helper.fit_gauss_1d(xr, chunk, sigma_factor=sig_factor)
            fitted.mean_0.value += delta_add
        elif fitform.name == "Moffat1D":
            fitted = math_helper.fit_moffat_1d(chunk, sigma_factor=sig_factor)
            fitted.x_0_0.value += delta_add
        elif fitform.name == "MexicanHat1D":
            fitted = math_helper.fit_mex_hat_1d(chunk, sigma_factor=sig_factor)
            fitted.x_0_0.value += delta_add
        elif fitform.name == "Polynomial1D":
            fitted = math_helper.fit_poly_n(chunk, deg=degree, sigma_factor=sig_factor)
            if fitted is None:
                raise ValueError("Problem with the Poly1D fit")
        elif fitform.name == "AiryDisk2D":
            fitted = math_helper.fit_airy_2d(chunk, sigma_factor=sig_factor)
            if fitted is None:
                raise ValueError("Problem with the AiryDisk2D fit")
            fitted.x_0_0.value += (xx - delta)
            fitted.y_0_0.value += (yy - delta)

        xline = np.arange(len(chunk)) + delta_add
        fline = np.linspace(xline[0], xline[-1], 100)  # finer sample

        if fitform.name == "AiryDisk2D":
            yfit = fitted(fline, fline * 0 + fitted.y_0_0.value)
        else:
            yfit = fitted(fline)

        # make a plot
        if pars["title"][0] is None:
            title = f"{self._datafile}: {int(x)} {int(y)}\n"
        else:
            title = pars["title"][0]

        if genplot:
            pfig = fig
            if fig is None:
                fig = plt.figure(self._figure_name)
            fig.clf()
            fig.add_subplot(111)
            ax = fig.gca()

            ax.set_xlabel(pars["xlabel"][0])
            ax.set_ylabel(pars["ylabel"][0])
            if pars["logx"][0]:
                ax.set_xscale("log")
            if pars["logy"][0]:
                ax.set_yscale("log")

            if bool(pars["pointmode"][0]):
                ax.plot(xline, chunk, 'o', label="data")
            else:
                ax.plot(xline, chunk, label="data", linestyle='-')

            if fitform.name == "Gaussian1D":
                fwhmx, fwhmy = math_helper.gfwhm(fitted.stddev_0.value)
                ax.set_title(f"{title} amp={fitted.amplitude_0.value:8.2f}"
                             f" mean={fitted.mean_0.value:9.2f},"
                             f"fwhm={fwhmx:9.2f}")
                pstr = f"({int(x):d},{int(y):d}) mean={fitted.mean_0.value:9.2f}, fwhm={fwhmx:9.2f}"
                self.log.info(pstr)
            elif fitform.name == "Moffat1D":
                mfwhm = math_helper.mfwhm(fitted.alpha_0.value,
                                          fitted.gamma_0.value)
                ax.set_title(f"{title} amp={fitted.amplitude_0.value:8.2f}"
                             f" fwhm={mfwhm:9.2f}")
                pstr = f"({int(x):d},{int(y):d}) amp={fitted.amplitude_0.value:8.2f} fwhm={mfwhm:9.2f}"
                self.log.info(pstr)
            elif fitform.name == "MexicanHat1D":
                ax.set_title(f"{title} amp={fitted.amplitude_0.value:8.2f} sigma={fitted.sigma_0.value:8.2f}")
                pstr = f"({int(x):d},{int(y):d}) amp={fitted.amplitude_0.value:8.2f} sigma={fitted.sigma_0.value:9.2f}"
                self.log.info(pstr)
            elif fitform.name == "Polynomial1D":
                ax.set_title(f"{title} degree={degree}")
                pstr = f"({int(x):d},{int(y):d}) degree={degree}"
                self.log.info(fitted.parameters)
                self.log.info(pstr)
            elif fitform.name == "AiryDisk2D":
                ax.set_title(f"{title} amp={fitted.amplitude_0.value:8.2f} radius={fitted.radius_0.value:8.2f}")
                pstr = f"({int(x):d},{int(y):d}) amp={fitted.amplitude_0.value:8.2f} radius={fitted.radius_0.value:9.2f}"
                self.log.info(pstr)
            else:
                warnings.warn("Unsupported functional form specified for fit")
                raise ValueError
            ax.plot(fline, yfit, c='r', label=str(fitform.__name__) + " fit")
            ax.legend()

            if pfig is None and 'nbagg' not in self._mpl_backend:
                plt.draw()
                plt.pause(0.001)
            else:
                fig.canvas.draw_idle()

        else:
            return fitted

    def column_fit(self, x, y, data=None, form=None, genplot=True, fig=None):
        """Compute the 1d fit to the column of data.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        form: string
            This is the functional form specified in the column fit parameters
        genplot: int
            produce the plot or return the fit model
        fig: figure name for redirect
            Used for interaction with the ginga GUI

        Notes
        -----
        delta is the range of data values to use around the x,y location

        The background is currently ignored

        if centering is True, then the center is fit with a 2d gaussian,
        but this is currently not done for Polynomial1D
        """

        result = self.line_fit(x, y, data=data, form=form,
                               genplot=genplot, fig=fig, col=True)
        if not genplot:
            return result

    def com_center(self, x, y, data=None, delta=None, oversampling=1.):
        """ Return the center of mass of the object at x,y

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        delta: int
            The range of data values (bounding box) to use around the x,y
            location for calculating the center
        oversampling: int
            Oversampling factors of pixel indices. If oversampling
            is a scalar this is treated as both x and y directions
            having the same oversampling factor; otherwise it is
            treated as (x_oversamp, y_oversamp)


        """
        if data is None:
            data = self._data

        if delta is None:
            delta = int(self.com_center_pars['delta'][0])

        # reset delta for small arrays
        if delta >= len(data) / 4:
            delta = delta // 2

        if delta is None:
            delta = int(self.com_center_pars['delta'][0])
        xx = int(x)
        yy = int(y)

        #  flipped from xpa
        chunk = data[yy - delta:yy + delta, xx - delta:xx + delta]

        try:
            xcenter, ycenter = centroid_com(chunk, oversampling=oversampling)

            pstr = f"xc={(xcenter + xx - delta):.4f}\tyc={(ycenter + yy - delta):.4f}"
        except AttributeError:
            raise AttributeError("Problem with center of mass")
        self.log.info(pstr)

        return (xcenter + xx - delta,
                ycenter + yy - delta)


    def gauss_center(self, x, y, data=None, delta=10,
                     sigma_factor=0):
        """Return the Gaussian 2D fit center of the object at (x,y).

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        delta: int
            The range of data values (bounding box) to use around the x,y
            location for calculating the center
        sigma_factor: float, optional
            The sigma clipping factor to use on the data fit

        """
        if data is None:
            data = self._data

        # reset delta for small arrays
        if delta >= len(data) / 4:
            delta = delta / 2

        delta = int(delta)
        xx = int(x)
        yy = int(y)

        #  flipped from xpa
        chunk = data[yy - delta:yy + delta, xx - delta:xx + delta]
        try:
            fit = math_helper.fit_gaussian_2d(chunk, sigma_factor=sigma_factor)
            amp = fit.amplitude_0.value
            xcenter = fit.x_mean_0.value
            ycenter = fit.y_mean_0.value
            xsigma = fit.x_stddev_0.value
            ysigma = fit.y_stddev_0.value

            pstr = f"xc={(xcenter + xx - delta):.4f}\tyc={(ycenter + yy - delta):.4f}"
            self.log.info(pstr)
            return (amp,
                    xcenter + xx - delta,
                    ycenter + yy - delta,
                    xsigma,
                    ysigma)

        except (RuntimeError, UserWarning) as e:
            self.log.info(f"Warning: {str(e)}, returning zeros for fit")
            return (0, 0, 0, 0, 0)

    def radial_profile(self, x, y, data=None, form=None,
                       genplot=True, fig=None):
        """Display the radial profile plot (intensity vs radius) for the object.

        From the parameters Dictionary:
        If pixel is True, then every pixel at each radius is plotted.
        If pixel is False, then the sum of all pixels in integer bins is plotted

        Background may be subtracted and centering can be done with a
        2D Gaussian fit. These options are read from the plot parameters dict.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        form: string
            The string name of the form of the fit to use
        genplot: bool
            Generate the plot if True, else retfurn the fit data

        """
        pars = self.radial_profile_pars

        subtract_background = bool(pars["background"][0])
        if not photutils_installed and subtract_background:
            self.log.warning("Install photutils to enable "
                             "background subtraction")
            subtract_background = False

        if data is None:
            data = self._data

        getdata = bool(pars["getdata"][0])
        center = pars["center"][0]

        # be careful with the clipping since most
        # of the data will be near the low value
        clip_on = pars["clip"][0]
        if clip_on:
            sig_factor = pars["sigma"][0]
        else:
            sig_factor = 0

        fitplot = bool(pars["fitplot"][0])

        if fitplot:
            if form is None:
                fitform = getattr(models, pars["func"][0])
            else:
                if form not in self._fit_models:
                    msg = f"{form} not supported for fitting"
                    self.log.info(msg)
                    raise ValueError(msg)
                else:
                    fitform = getattr(models, form)

        # cut the data down to size and center cutout
        datasize = int(pars["rplot"][0])
        if datasize < 3:
            self.log.info("Insufficient pixels, resetting chunk size to 3.")
            datasize = 3

        if center:
            # reset delta for small arrays
            # make it odd if it's even
            if ((datasize % 2) == 0):
                datasize = datasize + 1
            xx = int(x)
            yy = int(y)
            #  flipped from xpa
            data_chunk = data[yy - datasize:yy + datasize,
                              xx - datasize:xx + datasize]
            amp, centerx, centery, sigmax, sigmay = self.gauss_center(xx, yy, data, delta=datasize)

        else:
            centery = y
            centerx = x

        icenterx = int(centerx)
        icentery = int(centery)

        # fractional center, help with precision errors to 1000th pixel
        xfrac = round(centerx - icenterx, 2)
        yfrac = round(centery - icentery, 2)

        # just grab the data box centered on the object
        data_chunk = data[icentery - datasize:icentery + datasize,
                          icenterx - datasize:icenterx + datasize]

        y, x = np.indices(data_chunk.shape)  # index of all pixels
        y = y - datasize
        x = x - datasize
        r = np.sqrt((x-xfrac)**2 + (y-yfrac)**2)

        indices = np.argsort(r.flat)  # sorted indices

        if pars["pixels"][0]:
            flux = data_chunk.ravel()[indices]
            radius = r.ravel()[indices]

        else:  # sum the flux in integer bins
            radius = r.ravel()[indices].astype(np.int)
            flux = np.bincount(radius, data_chunk.ravel()[indices])
            radbc = np.bincount(radius)
            flux = flux / radbc
            radius = np.arange(len(flux))

        # Get a background measurement
        if subtract_background:
            inner = pars["skyrad"][0]
            width = pars["width"][0]
            annulus_apertures = photutils.CircularAnnulus((centerx, centery),
                                                          r_in=inner,
                                                          r_out=inner+width)
            bkgflux_table = photutils.aperture_photometry(data,
                                                          annulus_apertures)

            # to calculate the mean local background, divide the circular
            # annulus aperture sums by the area of the circular annulus.
            # The bkg sum with the circular aperture is then
            # the mean local background times the circular apreture area.
            if photutils_version >= version.parse('0.7'):
                annulus_area = annulus_apertures.area
            else:
                annulus_area = annulus_apertures.area()
            sky_per_pix = float(bkgflux_table['aperture_sum'] /
                                annulus_area)

            # don't add flux
            if sky_per_pix < 0.:
                sky_per_pix = 0.
                self.log.info("Sky background negative, setting to zero")
            self.log.info(f"Background per pixel: {sky_per_pix}")
            flux -= sky_per_pix

            if getdata:
                self.log.info(f"Sky per pixel: {sky_per_pix} using "
                              f"(rad={inner}->{inner + width})")
        if getdata:
            info = f"\nat (x,y)={centerx},{centery}\n"
            self.log.info(info)
            self.log.info(radius, flux)

        # Fit the functional form to the radial profile flux
        # TODO: Ignore sky subtracted pixels that push flux
        # below zero?
        if fitplot:
            fline = np.linspace(0, datasize, 100)  # finer sample
            # fit model to data
            if fitform.name == "Gaussian1D":
                fitted = math_helper.fit_gauss_1d(radius, flux,
                                                  sigma_factor=sig_factor,
                                                  center_at=0,
                                                  weighted=True)

                fwhmx, fwhmy = math_helper.gfwhm(fitted.stddev_0.value)
                legend = (f"Max. pix. flux = {np.max(flux):9.3f}\n"
                          f"amp = {fitted.amplitude_0.value:9.3f}\n"
                          f"fwhm = {fwhmx:9.3f}")
                self.log.info(legend)
                legendx = datasize / 2
                legendy = np.max(flux) / 2

            elif fitform.name == "Moffat1D":
                fitted = math_helper.fit_moffat_1d(flux,
                                                   sigma_factor=sig_factor,
                                                   center_at=0,
                                                   weighted=True)
                mfwhm = math_helper.mfwhm(fitted.alpha_0.value,
                                          fitted.gamma_0.value)
                legend = (f"Max. pix. flux = {np.max(flux):9.3f}\n"
                          f"amp = {fitted.amplitude_0.value:9.3f}\n"
                          f"fwhm = {mfwhm:9.3f}")
                legendx = datasize / 2
                legendy = np.max(flux) / 2

            elif fitform.name == "MexicanHat1D":
                fitted = math_helper.fit_mex_hat_1d(flux,
                                                    sigma_factor=sig_factor,
                                                    center_at=0,
                                                    weighted=True)
                legend = (f"Max. pix. flux = {np.max(flux):9.3f}\n")
                legendx = datasize / 2
                legendy = np.max(flux) / 2

            if fitted is None:
                msg = f"Problem with the {fitform.name} fit"
                self.log.info(msg)
                raise ValueError(msg)

            yfit = fitted(fline)

        #  finish the plot
        #  TODO: users may get error if they use this without a display
        #  and request data back but forget to set genplot=False
        if genplot:
            pfig = fig
            if fig is None:
                fig = plt.figure(self._figure_name)
            fig.clf()
            fig.add_subplot(111)
            ax = fig.gca()

            if subtract_background:
                ytitle = (f"Flux ( sky/pix = {sky_per_pix:8.2f} )")
            else:
                ytitle = pars["ylabel"][0]
            ax.set_xlabel(pars["xlabel"][0])
            ax.set_ylabel(ytitle)

            if bool(pars["pointmode"][0]):
                ax.plot(radius, flux, pars["marker"][0])
            else:
                ax.plot(radius, flux)
            ax.set_ylim(0,)

            if pars["title"][0] is None:
                if fitplot:
                    title = f"Radial Profile at ({icenterx},{icentery}) with {fitform.name}"
                else:
                    title = f"Radial Profile for {icenterx} {icentery}"
            else:
                title = pars["title"][0]

            if fitplot:
                ax.plot(fline, yfit, linestyle='-', c='r', label=fitform.name)
                ax.set_xlim(0, datasize, 0.5)
                ax.text(legendx, legendy, legend)

            ax.set_title(title)

            if pfig is None and 'nbagg' not in self._mpl_backend:
                plt.draw()
                plt.pause(0.001)
            else:
                fig.canvas.draw_idle()
        else:
            return radius, flux

    def curve_of_growth(self, x, y, data=None, genplot=True, fig=None):
        """Display a curve of growth plot.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        fig: figure name for redirect
            Used for interaction with the ginga GUI

        Notes
        -----
        the object photometry is taken from photutils

        """
        if not photutils_installed:
            self.log.warning("Install photutils to enable")
        else:

            if data is None:
                data = self._data

            delta = 10  # chunk size to find center
            subpixels = 10  # for line fit later

            # center using a center of mass
            if self.curve_of_growth_pars["center"][0]:
                if self.aper_phot_pars["center_com"][0]:
                    # use the center of mass
                    centerx, centery = self.com_center(x, y, delta=delta)
                else:
                    # user the gaussian2d
                    amp, centerx, centery, sigma, sigmay = \
                    self.gauss_center(x, y, data, delta=delta)
            else:
                centery = y
                centerx = x
            centerx = int(centerx)
            centery = int(centery)

            # now grab aperture sums going out from that central pixel
            inner = self.curve_of_growth_pars["buffer"][0]
            width = self.curve_of_growth_pars["width"][0]
            router = self.curve_of_growth_pars["rplot"][0]
            getdata = bool(self.curve_of_growth_pars["getdata"][0])

            radius = list()
            flux = list()
            rapert = int(router) + 1
            for rad in range(1, rapert, 1):
                aper_flux, annulus_sky, skysub_flux = self._aperture_phot(
                    centerx, centery, data, radsize=rad, sky_inner=inner,
                    skywidth=width, method="exact", subpixels=subpixels)
                radius.append(rad)
                if self.curve_of_growth_pars["background"][0]:
                    if inner < router:
                        warnings.warn(
                            "Your sky annulus is inside your \
                            photometry radius rplot")
                    flux.append(skysub_flux)
                else:
                    flux.append(aper_flux)
            if getdata:
                rapert = np.arange(1, rapert, 1)
                info = f"\nat (x,y)={int(centerx)},{int(centery)}\nradii:{rapert}\nflux:{flux}"
                self.log.info(info)
            if genplot:
                pfig = fig
                if fig is None:
                    fig = plt.figure(self._figure_name)
                fig.clf()
                fig.add_subplot(111)
                ax = fig.gca()

                if self.curve_of_growth_pars["title"][0] is None:
                    title = f"{self._datafile}: {int(x)} {int(y)}\n"
                else:
                    title = self.curve_of_growth_pars["title"][0]

                ax.set_xlabel(self.curve_of_growth_pars["xlabel"][0])
                ax.set_ylabel(self.curve_of_growth_pars["ylabel"][0])
                ax.plot(radius, flux, 'o')
                ax.set_title(title)

                if pfig is None and 'nbagg' not in self._mpl_backend:
                    plt.draw()
                    plt.pause(0.001)
                else:
                    fig.canvas.draw_idle()

            else:
                return rapert, flux

    def _aperture_phot(self, x, y, data=None, radsize=1,
                       sky_inner=5, skywidth=5,
                       method="subpixel", subpixels=4):
        """Perform sky subtracted aperture photometry.

        uses photutils functions, photutil must be installed

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        radsize: int
            Size of the radius
        sky_inner: int
            Inner radius of the sky annulus
        skywidth: int
            Width of the sky annulus
        method: string
            Pixel sampling method to use
        subpixels: int
            How many subpixels to use

        Notes
        -----
        background is taken from sky annulus pixels, check into
        masking bad pixels
        """
        if not photutils_installed:
            self.log.warning("Install photutils to enable")
        else:

            if data is None:
                data = self._data

            apertures = photutils.CircularAperture((x, y), radsize)
            rawflux_table = photutils.aperture_photometry(
                data,
                apertures,
                subpixels=1,
                method="center")

            outer = sky_inner + skywidth
            annulus_apertures = photutils.CircularAnnulus(
                (x, y), r_in=sky_inner, r_out=outer)
            bkgflux_table = photutils.aperture_photometry(
                data,
                annulus_apertures)

            # to calculate the mean local background, divide the circular
            # annulus aperture sums
            # by the area of the circular annulus. The bkg sum within the
            # circular aperture is then
            # then mean local background times the circular apreture area.
            if photutils_version >= version.parse('0.7'):
                aperture_area = apertures.area
                annulus_area = annulus_apertures.area
            else:
                aperture_area = apertures.area()
                annulus_area = annulus_apertures.area()
            bkg_sum = (
                bkgflux_table['aperture_sum'] *
                aperture_area /
                annulus_area)[0]
            skysub_flux = rawflux_table['aperture_sum'][0] - bkg_sum

            return (
                float(rawflux_table['aperture_sum'][0]), bkg_sum, skysub_flux)

    def histogram(self, x, y, data=None, genplot=True, fig=None):
        """Calulate a histogram of the data values.

        Parameters
        ----------
        x: int, required
            The x location of the object
        y: int, required
            The y location of the object
        data: numpy array, optional
            The data array to work on
        genplot: boolean, optional
            If false, returns the hist, bin_edges tuple
        fig: figure name for redirect
            Used for interaction with the ginga GUI

        Notes
        -----
        This functional originally used the pylab histogram routine for
        plotting. In order to accomodate returning just the histogram data,
        this was changed to the numpy histogram, with a subsequent plot if
        genplot is True.

        Does not yet support numpy v1.11 strings for bin estimation.

        """
        if data is None:
            data = self._data

        deltax = int(self.histogram_pars["ncolumns"][0] / 2.)
        deltay = int(self.histogram_pars["nlines"][0] / 2.)
        yf = int(y)
        xf = int(x)

        data_cut = data[yf - deltay:yf + deltay, xf - deltax:xf + deltax]

        # mask data for min and max intensity specified
        if self.histogram_pars["z1"][0]:
            mini = float(self.histogram_pars["z1"][0])
        else:
            mini = np.min(data_cut)

        if self.histogram_pars["z2"][0]:
            maxi = float(self.histogram_pars["z2"][0])
        else:
            maxi = np.max(data_cut)

        lt = (data_cut < maxi)
        gt = (data_cut > mini)

        total_mask = lt * gt
        flat_data = data_cut[total_mask].flatten()

        if not maxi:
            maxi = np.max(data_cut)
        if not mini:
            mini = np.min(data_cut)
        num_bins = int(self.histogram_pars["nbins"][0])

        if genplot:
            pfig = fig
            if fig is None:
                fig = plt.figure(self._figure_name)
            fig.clf()
            fig.add_subplot(111)
            ax = fig.gca()

            if self.histogram_pars["title"][0] is None:
                title = f"{self._datafile}: {int(x)} {int(y)}\n"
            else:
                title = self.histogram_pars["title"][0]
            ax.set_title(title)
            ax.set_xlabel(self.histogram_pars["xlabel"][0])
            ax.set_ylabel(self.histogram_pars["ylabel"][0])

            if self.histogram_pars["logx"][0]:
                ax.set_xscale("log")
            if self.histogram_pars["logy"][0]:
                ax.set_yscale("log")
            n, bins, patches = ax.hist(flat_data, num_bins,
                                       range=[mini, maxi],
                                       density=False,
                                       facecolor='green',
                                       alpha=0.5,
                                       histtype='bar')
            self.log.info(f"{num_bins} bins "
                          f"range:[{mini},{maxi}]")

            if pfig is None and 'nbagg' not in self._mpl_backend:
                plt.draw()
                plt.pause(0.001)
            else:
                fig.canvas.draw_idle()
        else:
            hist, bin_edges = np.histogram(flat_data,
                                           num_bins,
                                           range=[mini, maxi],
                                           density=False)
            return hist, bin_edges

    def contour(self, x, y, data=None, fig=None):
        """plot contours in a region around the specified location.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        fig: figure for redirect
            Used for interaction with the ginga GUI

        """
        if data is None:
            data = self._data

        pfig = fig
        if fig is None:
            fig = plt.figure(self._figure_name)
        fig.clf()
        fig.add_subplot(111)
        ax = fig.gca()

        if self.contour_pars["title"][0] is None:
            title = f"{self._datafile} {int(x)} {int(y)}"
        else:
            title = self.contour_pars["title"][0]
        ax.set_title(title)
        ax.set_xlabel(self.contour_pars["xlabel"][0])
        ax.set_ylabel(self.contour_pars["ylabel"][0])
        ncont = self.contour_pars["ncontours"][0]
        colormap = self.contour_pars["cmap"][0]
        lsty = self.contour_pars["linestyles"][0]

        self.log.info(f"contour centered at: {x} {y}")
        deltax = int(self.contour_pars["ncolumns"][0] / 2.)
        deltay = int(self.contour_pars["nlines"][0] / 2.)
        xx = int(x)
        yy = int(y)
        data_cut = data[yy - deltay:yy + deltay, xx - deltax:xx + deltax]
        plt.rcParams['xtick.direction'] = 'out'
        plt.rcParams['ytick.direction'] = 'out'

        X, Y = np.meshgrid(np.arange(0, deltax, 0.5) + x - deltax / 2.,
                           np.arange(0, deltay, 0.5) + y - deltay / 2.)

        C = ax.contour(
            X,
            Y,
            data_cut,
            ncont,
            linewidths=.5,
            colors='black',
            linestyles=lsty)
        # make the filled contour
        ax.contourf(X, Y, data_cut, ncont, alpha=.75, cmap=colormap)
        if self.contour_pars["label"][0]:
            ax.clabel(C, inline=1, fontsize=10, fmt="%5.3f")

        if pfig is None and 'nbagg' not in self._mpl_backend:
            plt.draw()
            plt.pause(0.001)
        else:
            fig.canvas.draw_idle()

    def surface(self, x, y, data=None, fig=None):
        """plot a surface around the specified location.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        fig: figure for redirect
            Used for interaction with the ginga GUI
        """
        from mpl_toolkits.mplot3d import Axes3D
        from matplotlib.ticker import LinearLocator, FormatStrFormatter

        if data is None:
            data = self._data
        pfig = fig
        if fig is None:
            fig = plt.figure(self._figure_name)
        fig.clf()
        fig.add_subplot(111)
        ax = fig.gca(projection='3d')

        title = self.surface_pars["title"][0]
        if title is None:
            title = f"{self._datafile}: {int(x)} {int(y)}"
        ax.set_title(title)
        ax.set_xlabel(self.surface_pars["xlabel"][0])
        ax.set_ylabel(self.surface_pars["ylabel"][0])
        if self.surface_pars["zlabel"][0]:
            ax.set_zlabel("Flux")
        fancy = self.surface_pars["fancy"][0]
        deltax = self.surface_pars["ncolumns"][0]
        deltay = self.surface_pars["nlines"][0]

        minx = int(x - deltax) if x - deltax > 0 else 0
        miny = int(y - deltay) if y - deltay > 0 else 0
        maxx = int(x + deltax) if x + deltax < data.shape[-1] else data.shape[0]
        maxy = int(y + deltay) if y + deltay < data.shape[0] else data.shape[0]

        X = np.arange(minx, maxx, 1)
        Y = np.arange(miny, maxy, 1)

        X, Y = np.meshgrid(X, Y)
        Z = data[miny:maxy, minx:maxx]
        if self.surface_pars["floor"][0]:
            zmin = float(self.surface_pars["floor"][0])
        else:
            zmin = np.min(Z)
        if self.surface_pars["ceiling"][0]:
            zmax = float(self.surface_pars["ceiling"][0])
        else:
            zmax = np.max(Z)

        stride = int(self.surface_pars["stride"][0])
        if fancy:
            surf = ax.plot_surface(
                X, Y, Z, rstride=stride, cstride=stride,
                cmap=self.surface_pars["cmap"][0], alpha=0.6)
        else:
            surf = ax.plot_surface(X, Y, Z, rstride=stride, cstride=stride,
                                   cmap=self.surface_pars["cmap"][0],
                                   linewidth=0, antialiased=False)

        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter(FormatStrFormatter('%.0f'))
        ax.set_zlim(zmin, zmax)

        if fancy:
            xmin = minx
            ymax = maxy
            cset = ax.contour(
                X,
                Y,
                Z,
                zdir='z',
                offset=zmax,
                cmap=self.surface_pars["cmap"][0])
            cset = ax.contour(
                X,
                Y,
                Z,
                zdir='x',
                offset=xmin,
                cmap=self.surface_pars["cmap"][0])
            cset = ax.contour(
                X,
                Y,
                Z,
                zdir='y',
                offset=ymax,
                cmap=self.surface_pars["cmap"][0])

        fig.colorbar(surf, shrink=0.5, aspect=5)
        if self.surface_pars["azim"][0]:
            ax.view_init(elev=10., azim=float(self.surface_pars["azim"][0]))

        if pfig is None and 'nbagg' not in self._mpl_backend:
            plt.draw()
            plt.pause(0.001)
        else:
            fig.canvas.draw_idle()

    def cutout(self, x, y, data=None, size=None, fig=None):
        """Make a fits cutout around the pointer location without wcs.

        Parameters
        ----------
        x: int
            The x location of the object
        y: int
            The y location of the object
        data: numpy array
            The data array to work on
        size: int
            The radius of the cutout region
        fig: figure for redirect
            Used for interaction with the ginga GUI
        """
        if data is None:
            data = self._data

        if size is None:
            size = self.cutout_pars["size"][0]

        xx = int(x)
        yy = int(y)
        prefix = f"cutout_{xx}_{yy}_"
        fname = tempfile.mkstemp(prefix=prefix, suffix=".fits", dir="./")[-1]
        cutout = data[yy - size:yy + size, xx - size:xx + size]
        hdu = fits.PrimaryHDU(cutout)
        hdulist = fits.HDUList([hdu])
        hdulist[0].header['EXTEND'] = False
        hdulist.writeto(fname)
        self.log.info(f"Cutout at ({xx},{yy}) saved to {fname}")

    def register(self, user_funcs):
        """register a new imexamine function made by the user as an option.

        Parameters
        ----------
        user_funcs: dict
            Contains a dictionary where each key is the binding for the
            (function,description) tuple

        Notes
        -----
        The new binding will be added to the dictionary of imexamine functions
        as long as the key is unique. The new functions do not have to have
        default dictionaries associated with them.
        """
        if not isinstance(user_funcs, type(dict())):
            warnings.warn("Your input needs to be a dictionary")

        for key in user_funcs.keys():
            if key in self.imexam_option_funcs.keys():
                warnings.warn(f"{key} is not a unique key")
                warnings.warn(f"{self.imexam_option_funcs[key]}")
                raise ValueError(f"{key} is not a unique key")
            elif key == 'q':
                warnings.warn("q is reserved as the quit key")
                raise ValueError("q is reserved for the quit key")
            else:
                func_name = user_funcs[key][0].__name__
                self._add_user_function(user_funcs[key][0])
                self.imexam_option_funcs[key] = (
                    self.__getattribute__(func_name), user_funcs[key][1])
                self.log.info(f"User function: {func_name} added to imexam options with "
                    f"key {key}")

    @classmethod
    def _add_user_function(cls, func):
        import types
        if PY3:
            return setattr(cls, func.__name__, types.MethodType(func, cls))
        else:
            return setattr(cls, func.__name__,
                           types.MethodType(func, None, cls))

    def set_aper_phot_pars(self, user_dict=None):
        """the user may supply a dictionary of par settings."""
        if not user_dict:
            self.aper_phot_pars = imexam_defpars.aper_phot_pars
        else:
            self.aper_phot_pars = user_dict

    def set_com_center_pars(self):
        """ set paramters for the center of mass function"""
        self.com_center_pars = imexam_defpars.radial_profile_pars

    def set_radial_pars(self):
        """set parameters for radial profile plots."""
        self.radial_profile_pars = imexam_defpars.radial_profile_pars

    def set_curve_pars(self):
        """set parameters for curve of growth plots."""
        self.curve_of_growth_pars = imexam_defpars.curve_of_growth_pars

    def set_surface_pars(self):
        """set parameters for surface plots."""
        self.surface_pars = imexam_defpars.surface_pars

    def set_line_fit_pars(self):
        """set parameters for 1D line fit plots."""
        self.line_fit_pars = imexam_defpars.line_fit_pars

    def set_column_fit_pars(self):
        """set parameters for 1D line fit plots."""
        self.column_fit_pars = imexam_defpars.column_fit_pars

    def set_contour_pars(self):
        """set parameters for contour plots."""
        self.contour_pars = imexam_defpars.contour_pars

    def set_histogram_pars(self):
        """set parameters for histogram plots."""
        self.histogram_pars = imexam_defpars.histogram_pars

    def set_lineplot_pars(self):
        """set parameters for line plots."""
        self.lineplot_pars = imexam_defpars.lineplot_pars

    def set_colplot_pars(self):
        """set parameters for column plots."""
        self.colplot_pars = imexam_defpars.colplot_pars

    def set_cutout_pars(self):
        """set parameters for cutout images."""
        self.cutout_pars = imexam_defpars.cutout_pars

    def reset_defpars(self):
        """set all pars to their defaults."""
        self._define_pars()



"""Licensed under a 3-clause BSD style license - see LICENSE.rst."""

import os
import sys
import logging
import warnings
from astropy.io import fits

from . import __version__

try:
    from . import imexamxpa as xpa
    _have_xpa = True
except ImportError:
    _have_xpa = False


# To guide any import *
__all__ = ["display_help", "set_logging"]


def find_path(target=None):
    """Find the local path to the target executable.

    Parameters
    ----------
    target: None or str
        The name of the executable to find

    Returns
    -------
    The path to the executable or None if not found

    Notes
    -----
    This will first look for the target on the users
    path and then look to see if an alias has been
    set. It assumes that the user has aliased ds9 to
    the same name, ds9. If otherwise, the user should
    specify path in the call to the connect() method.
    """

    if target is None:
        raise TypeError("Expected name of target executable")

    found_path = None
    for dirname in os.getenv("PATH").split(":"):
        possible = os.path.join(dirname, target)
        if os.path.isfile(possible):
            found_path = possible
    if found_path is None:
        # This will also return None if not set
        found_path = os.getenv(target)
    return found_path


def list_active_ds9(verbose=True):
    """
    Display and/or return information about the DS9 windows currently
    registered with the XPA.

    Parameters
    ----------
    verbose : bool
        If True, prints out all the information about what DS9 windows
        are active.

    Returns
    -------
    session_list : list
        The list of sessions that have been registered.  Each entry in the list
        is a list containing the information that xpans yields.  Typically the
        fourth element in that tuple contains the actual target name.

    Notes
    -----
    when I start a unix socket with connect() the xpa register isn't
    seeing it when I call this function. I think because it's only
    listening on the inet socket which starts by default in the OS.
    That's if xpans is installed on the machine. Otherwise, no
    nameserver is running at all.
    """
    session_dict = {}

    # only run if XPA/xpans is installed on the machine
    if find_path('xpans'):
        sessions = None
        try:
            sessions = xpa.get(b"xpans").decode().strip().split("\n")
            if ((sessions is None or len(sessions) < 1) and verbose):
                print("No active sessions")
            else:
                for line in sessions:
                    classn, name, access, ids, user = tuple(line.split())
                    session_dict[ids] = (name, user, classn, access)
                if verbose:
                    for line in sessions:
                        print(line)
        except (ValueError, xpa.XpaException):
                print("No active sessions registered")

    else:
        print("XPA nameserver not installed or not on PATH, \
               function unavailable")

    return session_dict


def display_help():
    """Display RTD html help for the installed verison in a browser window."""
    url = "http://imexam.readthedocs.io/"
    try:
        import webbrowser
        # grab the version that's installed
        if "dev" not in __version__:
            url += f"en/v{__version__}/"
        webbrowser.open(url)
    except ImportError:
        warnings.warn(
            f"webbrowser module not installed, see {url} for help")
        raise ImportError


def display_xpa_help():
    """Display help for XPA Acccess Points."""
    url = "http://ds9.si.edu/doc/ref/xpa.html"
    try:
        import webbrowser
        webbrowser.open(url)
    except ImportError:
        warnings.warn(
            f"webbrowser module not installed, see {url} for help")
        raise ImportError


# Set up logging ability for the user
# consider making a private logging level for data retension
def set_logging(filename=None, on=True, level=logging.INFO):
    """Turn on or off logging to file or stdout.

    Parameters
    ----------

    filename: str, optional
        name of the file for logging information
    on: bool, optional
        turn logging on or off, will close the file output
        in order to turn off stdout logging, set the level
        to logging.CRITICAL
    level: logging, optional
        the logging level to be recorded or displayed

    """

    formatter = logging.Formatter('\n%(funcName)s \n%(message)s')
    root = logging.getLogger(__name__)
    root.setLevel(level)
    stream_attached = False
    file_attached = False

    if on:
        #  Try to avoid adding duplicate file handlers
        if len(root.handlers) > 0:
            for handler in root.handlers:
                if isinstance(handler, logging.StreamHandler):
                    stream_attached = True
                    handler.setLevel(level)
                if isinstance(handler, logging.FileHandler):
                    file_attached = True
                    raise ValueError("File for logging already specified,\
                                      turn off logging first.")
        else:
            # to prevent warning in unhandled contexts and messages to stderr
            root.addHandler(logging.NullHandler())

        if isinstance(filename, str) and not file_attached:
            file_handler = logging.FileHandler(filename=filename,
                                               mode='a',
                                               delay=True)
            file_handler.setLevel(logging.INFO)
            file_handler.setFormatter(formatter)
            print(f"Saving imexam commands to {repr(filename)}")
            root.addHandler(file_handler)

        if not stream_attached:
            # set the stdout stream handler
            stdout_handler = logging.StreamHandler(stream=sys.stdout)
            stdout_handler.setLevel(logging.INFO)
            # stdout_handler.setFormatter(formatter)
            root.addHandler(stdout_handler)

    #  turning the logging off to the file and set level on stream handler
    else:
        for handler in root.handlers:
            # close the file logger
            if isinstance(handler, logging.FileHandler):
                handler.close()
                root.removeHandler(handler)
            # set stream logging level
            if isinstance(handler, logging.StreamHandler):
                handler.setLevel(level)
    return root


def check_valid(fits_data=None):
    """Check the file to see if it is a multi-extension FITS file
    or a simple fits image where the data and header are stored in
    the primary header unit.

    Parameters
    ----------
    fits_data: None, FITS object
        Set to an in-memory FITS object if passing FITS HDUList;
        Otherwise set to the name of the file to check

    Returns
    -------
    mef_file: bool
        Returns True if the file is a multi-extension fits file
    nextend: int
        The number of extension in the file
    first_image: int, None
        The extension that contains the first image data.
        None will be returned when no IMAGE xtension is found

    Notes
    -----
    Drizzled images put a table in the first extension and an image in
    the zero extension, so this function checks for the first occurrance
    of 'IMAGE' in 'XTENSION', which is a required keyword.
    """
    log = logging.getLogger(__name__)
    found_image = False  # Does it contain an IMAGE XTENSION
    nextend = 0  # how many extenions does it have
    first_image = None  # what extension has the first image?
    fits_file = False
    mef_file = False

    if fits_data is None:
            raise ValueError("No filename or FITS object provided")
    if isinstance(fits_data, fits.hdu.hdulist.HDUList):
        mef_file = True
        fits_image = fits_data
    elif isinstance(fits_data, fits.hdu.image.PrimaryHDU):
        if fits_data.header['NAXIS'] > 0:
            first_image = 0
        fits_image = fits_data
    elif isinstance(fits_data, str):
        fits_file = True
        try:
            fits_image = fits.open(fits_data)
        except IOError:
            msg = f"Error opening file {repr(fits_data)}"
            log.warning(msg)
            raise IOError(msg)
        try:
            # EXTEND is required for MEF FITS files
            mef_file = fits_image[0].header['EXTEND']
            if not mef_file:
                if fits_image[0].header['NAXIS'] > 0:
                    first_image = 0
        except KeyError:
            if fits_image[0].header['NAXIS'] > 0:
                first_image = 0

    #  double check for lying liars, should at least have 1 extension
    #  if it's MEF and XTENSION is a required keyword in each extension
    #  after the primary hdu
    if mef_file:
        for extn in fits_image:
            try:
                if ((extn.header['XTENSION'] == 'IMAGE') and (not found_image)):
                    first_image = nextend  # The number of the extension
                    found_image = True
            except KeyError:
                # There doens't have to be an 'XTENSION' keyword in the global(0) if
                # the MEF has data there, so check for naxis if its an image
                # and tfields if it's table data
                if nextend == 0:
                    try:
                        tfields = extn.header['TFIELDS']  # table?
                    except KeyError:
                        if extn.header['NAXIS'] > 0:
                            found_image = True
                            first_image = nextend
                else:
                    mef_file = False
            nextend += 1
        nextend -= 1  # account for overcounting in return value
    if fits_file:
        fits_image.close()
    return (mef_file, nextend, first_image)


def verify_filename(filename=None, extver=None, extname=None):
    """Verify the filename exists and split it for extension information.
    If the user has given an extension, extension name or some combination of
    those, return the full filename, extension and version tuple.

    Parameters
    ----------
    filename: string
        The name of the file to verify

    extver: int
        extsion number which corresponds to extname

    extname: string
        the name of the extension

    Returns
    -------
    shortname: str
        The name of the file with full path
    extname: str, None
        The extension name, or None
    extver: int, None
        The extension version number or None
    """
    if filename is None:
        print("No filename provided")
    elif not isinstance(filename, str):
        raise TypeError("Expected filename to be a string")
    else:
        if "[" in filename:
            splitstr = filename.split("[")
            shortname = os.path.abspath(splitstr[0])
            if "," in splitstr[1]:
                extname = splitstr[1].split(",")[0]
                extver = int(splitstr[1].split(",")[1][0])
            else:
                extver = int(filename.split("[")[1][0])
        else:
            shortname = os.path.abspath(filename)
    return shortname, extname, extver
